"""
Viewer data routes

Routes:
    CONVERT_TIMEPOINTS: Convert timepoints to seconds
    GET_CLICK_COORDS: Get click coords
    GET_COORD_LABELS: Get coordinate labels
    GET_CROSSHAIR_COORDS: Get crosshair coords
    GET_DIRECTION_LABEL_COORDS: Get direction label coords
    GET_DISTANCE_DATA: Get distance data
    GET_FMRI_DATA: Get FMRI data
    GET_LAST_TIMECOURSE: Get last added fmri timecourse
    GET_MONTAGE_DATA: Get montage data
    GET_TASK_CONDITIONS: Get task conditions
    GET_TIMECOURSE_DATA: Get timecourse data
    GET_TIMECOURSE_LABELS: Get timecourse labels
    GET_TIMECOURSE_SOURCE: Get timecourse source
    GET_TIMEPOINT: Get timepoint
    GET_TIMEPOINTS: Get timepoints
    GET_VERTEX_COORDS: Get vertex coordinates
    GET_VIEWER_METADATA: Get viewer metadata
    GET_VOXEL_COORDS: Get voxel coordinates
    GET_WORLD_COORDS: Get world coordinates
    POP_FMRI_TIMECOURSE: Pop fmri timecourse
    REMOVE_FMRI_TIMECOURSES: Remove all fmri timecourses
    UPDATE_LOCATION: Update location
    UPDATE_FUNCTIONAL_TIMECOURSE: Update functional timecourse
    UPDATE_MONTAGE_SLICE_DIR: Update montage slice direction
    UPDATE_MONTAGE_SLICE_IDX: Update montage slice indices
    UPDATE_TIMEPOINT: Update timepoint
    UPDATE_TR: Update TR
"""
import json

from typing import List, Tuple
from flask import Blueprint, request

from findviz.logger_config import setup_logger
from findviz.routes.utils import (
    convert_value,
    handle_context, 
    handle_route_errors, 
    Routes
)
from findviz.routes.shared import data_manager
from findviz.routes.viewer.nifti import (
    get_nifti_data, get_timecourse_nifti
)
from findviz.routes.viewer.gifti import (
    get_gifti_data, get_timecourse_gifti
)

# Set up a logger for the app
logger = setup_logger(__name__)

data_bp = Blueprint('data', __name__)


@data_bp.route(Routes.CONVERT_TIMEPOINTS.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timepoints conversion request',
    log_msg='Timepoints conversion request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.CONVERT_TIMEPOINTS
)
def convert_timepoints() -> dict:
    """Convert timepoints to seconds"""
    data_manager.ctx.convert_timepoints()
    return {'status': 'success'}


@data_bp.route(Routes.GET_CLICK_COORDS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in click coords request',
    log_msg='Click coords request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_CLICK_COORDS
)
def get_click_coords() -> dict:
    """Get click coords"""
    return data_manager.ctx.get_click_coords()


@data_bp.route(Routes.GET_COORD_LABELS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in coordinate labels request',
    log_msg='Coordinate labels request successful',
    route=Routes.GET_COORD_LABELS
)
def get_coord_labels() -> dict:
    """Get coordinate labels"""
    if data_manager.ctx.fmri_file_type == 'nifti':
        return data_manager.ctx.coord_labels
    else:
        coord_labels = data_manager.ctx.coord_labels
        return {
            'left_coord_labels': coord_labels[0],
            'right_coord_labels': coord_labels[1]
        }


@data_bp.route(Routes.GET_CROSSHAIR_COORDS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in crosshair data request',
    log_msg='Crosshair data request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_CROSSHAIR_COORDS
)
def get_crosshair_coords() -> dict:
    """Get crosshair coordinates"""
    return data_manager.ctx.get_crosshair_coords()


@data_bp.route(Routes.GET_DIRECTION_LABEL_COORDS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in direction label coords request',
    log_msg='Direction label coords request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_DIRECTION_LABEL_COORDS
)
def get_direction_label_coords() -> dict:
    """Get direction label coordinates"""
    return data_manager.ctx.get_direction_label_coords()


@data_bp.route(Routes.GET_DISTANCE_DATA.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in distance data request',
    log_msg='Distance data request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_DISTANCE_DATA
)
def get_distance_data() -> list[float]:
    """Get distance data"""
    return data_manager.ctx.distance_data.tolist()


@data_bp.route(Routes.GET_FMRI_DATA.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in data update request',
    log_msg='Data update request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_FMRI_DATA
)
def get_fmri_data() -> dict:
    """Get FMRI data for the current timepoint and location."""
    # get plot options data from data manager
    plot_options = data_manager.ctx.get_fmri_plot_options()

    # get viewer data from data manager
    viewer_data = data_manager.ctx.get_viewer_data(
        fmri_data=True,
        time_course_data=False,
        task_data=False,
    )

    # pass viewer data to get_timepoint_data
    if data_manager.ctx.fmri_file_type == 'nifti':
        slice_idx = data_manager.ctx.get_slice_idx()
        timepoint_data = get_nifti_data(
            time_point=data_manager.ctx.timepoint,
            func_img=viewer_data['func_img'],
            coord_labels=data_manager.ctx.coord_labels,
            slice_idx=slice_idx,
            view_state=data_manager.ctx.view_state,
            montage_slice_dir=data_manager.ctx.montage_slice_dir,
            threshold_min=plot_options['threshold_min'],
            threshold_max=plot_options['threshold_max'],
            threshold_min_orig = data_manager.ctx.color_options_original['threshold_min'],
            threshold_max_orig = data_manager.ctx.color_options_original['threshold_max'],
            anat_img=viewer_data['anat_img']
        )
    else:
        timepoint_data = get_gifti_data(
            time_point=data_manager.ctx.timepoint,
            left_func_img=viewer_data['left_func_img'],
            right_func_img=viewer_data['right_func_img'],
            threshold_min=plot_options['threshold_min'],
            threshold_max=plot_options['threshold_max'],
            threshold_min_orig = data_manager.ctx.color_options_original['threshold_min'],
            threshold_max_orig = data_manager.ctx.color_options_original['threshold_max'],
        )

    return {
        'data': timepoint_data,
        'plot_options': plot_options
    }


@data_bp.route(Routes.GET_LAST_TIMECOURSE.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in last fmri timecourse request',
    log_msg='Last fmri timecourse request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_LAST_TIMECOURSE
)
def get_last_timecourse() -> dict:
    """Get last added fmri timecourse"""
    return data_manager.ctx.get_last_timecourse()


@data_bp.route(Routes.GET_MONTAGE_DATA.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in montage data request',
    log_msg='Montage data request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_MONTAGE_DATA
)
def get_montage_data() -> dict:
    """Get montage data for the current location."""
    montage_data = {
        'montage_slice_dir': data_manager.ctx.montage_slice_dir,
        'montage_slice_idx': data_manager.ctx.montage_slice_idx,
        'montage_slice_len': data_manager.ctx.slice_len
    }
    return montage_data


@data_bp.route(Routes.GET_TASK_CONDITIONS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in task conditions request',
    log_msg='Task conditions request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_TASK_CONDITIONS
)
def get_task_conditions() -> list[str]:
    """Get task conditions"""
    return data_manager.ctx.task_conditions


@data_bp.route(Routes.GET_TIMECOURSE_LABELS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timecourse labels request',
    log_msg='Timecourse labels request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_TIMECOURSE_LABELS
)
def get_timecourse_labels() -> list[str]:
    """Get timecourse labels"""
    return data_manager.ctx.ts_labels


@data_bp.route(Routes.GET_TIMECOURSE_LABELS_PREPROCESSED.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in preprocessed timecourse labels request',
    log_msg='Preprocessed timecourse labels request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_TIMECOURSE_LABELS_PREPROCESSED
)
def get_timecourse_labels_preprocessed() -> list[str]:
    """Get preprocessed timecourse labels"""
    return data_manager.ctx.ts_labels_preprocessed


@data_bp.route(Routes.GET_TIMECOURSE_DATA.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timecourse data request',
    log_msg='Timecourse data request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_TIMECOURSE_DATA,
    route_parameters=['ts_labels']
)
def get_timecourse_data() -> dict:
    """Get timecourse data for the current location."""
    ts_labels = json.loads(request.args['ts_labels'])
    viewer_data = data_manager.ctx.get_viewer_data(
        fmri_data=False,
        time_course_data=True,
        task_data=True,
    )
    timecourse_data = {}
    # add timecourse data if it exists
    if 'ts' in viewer_data:
        timecourse_data.update(viewer_data['ts'])
    # add task data if it exists
    if 'task' in viewer_data:
        timecourse_data.update(viewer_data['task'])
        
    # filter timecourse data to only include the requested ts_labels (if passed)
    if ts_labels is not None:
        timecourse_data = {
            ts_label: timecourse_data[ts_label]
            for ts_label in ts_labels
            if ts_label in timecourse_data
        }
    return timecourse_data


@data_bp.route(Routes.GET_TIMECOURSE_SOURCE.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timecourse source request',
    log_msg='Timecourse source request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_TIMECOURSE_SOURCE
)
def get_timecourse_source() -> dict:
    """Get timecourse source"""
    return {'timecourse_source': data_manager.ctx.timecourse_source}


@data_bp.route(Routes.GET_TIMEPOINT.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timepoint request',
    log_msg='Timepoint request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_TIMEPOINT
)
def get_timepoint() -> dict:
    """Get currently selected timepoint"""
    return {'timepoint': data_manager.ctx.timepoint}


@data_bp.route(Routes.GET_TIMEPOINTS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timepoints request',
    log_msg='Timepoints request successful',
    route=Routes.GET_TIMEPOINTS
)
def get_timepoints() -> dict:
    """Get all timepoints"""
    return {'timepoints': data_manager.ctx.get_timepoints()}


@data_bp.route(Routes.GET_VERTEX_COORDS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in vertex coordinates request',
    log_msg='Vertex coordinates request successful',
    route=Routes.GET_VERTEX_COORDS
)
def get_vertex_coords() -> dict:
    """Get vertex coordinates"""
    return {
        'vertex_number': data_manager.ctx.selected_vertex,
        'selected_hemisphere': data_manager.ctx.selected_hemi
    }

@data_bp.route(Routes.GET_VIEWER_METADATA.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in viewer metadata request',
    log_msg='Viewer metadata request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_VIEWER_METADATA
)
def get_viewer_metadata() -> dict:
    """Get viewer metadata"""
    return data_manager.ctx.get_viewer_metadata()


@data_bp.route(Routes.GET_VOXEL_COORDS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in voxel coordinates request',
    log_msg='Voxel coordinates request successful',
    route=Routes.GET_VOXEL_COORDS
)
def get_voxel_coords() -> dict:
    """Get voxel coordinates"""
    voxel_coords = data_manager.ctx.get_slice_idx()
    if data_manager.ctx.view_state == 'montage':
        voxel_coords = voxel_coords[data_manager.ctx.selected_slice]

    return voxel_coords


@data_bp.route(Routes.GET_WORLD_COORDS.value, methods=['GET'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in world coordinates request',
    log_msg='World coordinates request successful',
    route=Routes.GET_WORLD_COORDS
)
def get_world_coords() -> dict:
    world_coords = data_manager.ctx.get_world_coords()
    # package world coords into a dictionary
    world_coords_dict = {
        'x': world_coords[0],
        'y': world_coords[1],
        'z': world_coords[2]
    }
    return world_coords_dict


@data_bp.route(Routes.POP_FMRI_TIMECOURSE.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in fmri timecourse pop request',
    log_msg='Fmri timecourse pop request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.POP_FMRI_TIMECOURSE
)
def pop_fmri_timecourse() -> dict:
    """Pop fmri timecourse"""
    label = data_manager.ctx.pop_fmri_timecourse()
    return {'label': label}


@data_bp.route(Routes.REMOVE_FMRI_TIMECOURSES.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in fmri timecourse remove request',
    log_msg='Fmri timecourse remove request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.REMOVE_FMRI_TIMECOURSES
)
def remove_fmri_timecourses() -> dict:
    """Remove all fmri timecourses"""
    labels = data_manager.ctx.remove_fmri_timecourses()
    return {'labels': labels}


@data_bp.route(Routes.UPDATE_LOCATION.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in location update request',
    log_msg='Location update successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UPDATE_LOCATION,
    route_parameters=['click_coords', 'slice_name']
)
def update_location() -> dict:
    """Update current location based on form data."""
    click_coords = json.loads(request.form['click_coords'])
    slice_name = request.form['slice_name']
    data_manager.ctx.update_location(click_coords, slice_name)
    return {'status': 'success'}


@data_bp.route(Routes.UPDATE_FMRI_TIMECOURSE.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in fmri timecourse update request',
    log_msg='Fmri timecourse update request successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UPDATE_FMRI_TIMECOURSE
)
def update_fmri_timecourse() -> dict:
    """Update fmri timecourse data for the current location."""
    viewer_data = data_manager.ctx.get_viewer_data(
        fmri_data=True,
        time_course_data=False,
        task_data=False,
    )

    if data_manager.ctx.fmri_file_type == 'nifti':
        slice_idx = data_manager.ctx.get_slice_idx()
        # if montage view, get slice idx of currently selected slice
        if data_manager.ctx.view_state == 'montage':
            slice_idx = slice_idx[data_manager.ctx.selected_slice]
        timecourse_data, voxel_label = get_timecourse_nifti(
            func_img=viewer_data['func_img'],
            x=slice_idx['x'],
            y=slice_idx['y'],
            z=slice_idx['z']
        )
    else:
        click_coords = data_manager.ctx.get_click_coords()
        timecourse_data, voxel_label = get_timecourse_gifti(
            left_func_img=viewer_data['left_func_img'],
            right_func_img=viewer_data['right_func_img'],
            vertex_index=click_coords['selected_vertex'],
            hemisphere=click_coords['selected_hemi']
        )

    data_manager.ctx.update_timecourse(timecourse_data, voxel_label)
    return {'status': 'success'}


@data_bp.route(Routes.UPDATE_MONTAGE_SLICE_DIR.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in montage slice direction update request',
    log_msg='Montage slice direction update successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UPDATE_MONTAGE_SLICE_DIR,
    route_parameters=['montage_slice_dir']
)
def update_montage_slice_dir() -> dict:
    """Update montage slice direction"""
    montage_slice_dir = request.form['montage_slice_dir']
    data_manager.ctx.update_montage_slice_dir(montage_slice_dir)
    return {'status': 'success'}


@data_bp.route(Routes.UPDATE_MONTAGE_SLICE_IDX.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in montage slice indices update request',
    log_msg='Montage slice indices update successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UPDATE_MONTAGE_SLICE_IDX,
    route_parameters=['slice_name', 'slice_idx']
)
def update_montage_slice_idx() -> dict:
    """Update montage slice indices from slider changes"""
    slice_name = convert_value(request.form['slice_name'])
    slice_idx = convert_value(request.form['slice_idx'])
    data_manager.ctx.update_montage_slice_idx(slice_name, slice_idx)
    return {'status': 'success'}


@data_bp.route(Routes.UPDATE_TIMEPOINT.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in timepoint update request',
    log_msg='Timepoint update successful',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.UPDATE_TIMEPOINT,
    route_parameters=['time_point']
)
def update_timepoint() -> dict:
    """Update current timepoint based on form data."""
    timepoint = convert_value(request.form['time_point'])
    data_manager.ctx.update_timepoint(timepoint)
    return {'status': 'success'}


@data_bp.route(Routes.UPDATE_TR.value, methods=['POST'])
@handle_context()
@handle_route_errors(
    error_msg='Unknown error in TR update request',
    log_msg='TR update successful',
    route=Routes.UPDATE_TR
)
def update_tr() -> dict:
    """Update TR"""
    tr = convert_value(request.form['tr'])
    data_manager.ctx.set_tr(tr)
    return {'status': 'success'}



"""
Correlate fmri time courses with time course
"""
from typing import List

import numpy as np

from findviz.viz.analysis.utils import get_lag_mat
from findviz.viz.analysis.validate import (
    validate_less_than_or_equal_to_zero,
    validate_greater_than_or_equal_to_zero,
    validate_less_than_half_time_length
)
from findviz.viz.exception import ParameterInputError

class Correlate:
    def __init__(
        self, 
        negative_lag: int,
        positive_lag: int,
        time_length: int
    ):
        """
        Initialize Correlate class

        Parameters
        ----------
        negative_lag : int
            negative lag, should be zero or negative
        positive_lag : int
            positive lag, should be zero or positive
        time_length : int
            time length of fMRI data
        """
        self.negative_lag = negative_lag
        self.positive_lag = positive_lag
        self.time_length = time_length
        # validate parameters
        self._validate()
        # get array of lags
        self.lags = np.arange(
            self.negative_lag, self.positive_lag + 1
        )

    def correlate(self, fmri_data: np.ndarray, time_course: List[float]):
        """
        Correlate fmri data with time course

        Parameters
        ----------
        fmri_data : np.ndarray
            fmri data (n_timepoints, n_voxels)
        time_course : List[float]
            time course (n_timepoints)

        Returns
        -------
        corr : np.ndarray
            correlation map (n_lags, n_voxels)

        """
        # Convert time course to numpy array
        time_course = np.array(time_course)
        
        # Get lag matrix - shape (n_timepoints, n_lags)
        lag_mat = get_lag_mat(time_course[:, np.newaxis], self.lags)
        
        # Initialize correlation map
        n_lags = len(self.lags)
        n_voxels = fmri_data.shape[1]
        
        # Handle the case where fmri_data has 0 voxels
        if n_voxels == 0:
            return np.zeros((n_lags, 0))
        
        correlation_map = np.zeros((n_lags, n_voxels))
        
        # Compute correlations for each lag
        for i in range(n_lags):
            # Get the current lagged time course
            lagged_tc = lag_mat[:, i]
            
            # Find valid data points (not NaN)
            mask = ~np.isnan(lagged_tc)
            
            # Skip if not enough valid data points
            if np.sum(mask) <= 1:
                correlation_map[i, :] = np.nan
                continue
                
            # Use masked arrays to handle NaN values
            valid_lagged_tc = lagged_tc[mask]
            valid_fmri = fmri_data[mask, :]
                
            # Calculate correlation between lagged time course and all voxels at once
            # np.corrcoef returns a correlation matrix where:
            # - First row/column corresponds to the lagged time course
            # - Other rows/columns correspond to the voxels
            if valid_fmri.shape[1] > 0:  # Only if we have voxels
                corr_matrix = np.corrcoef(valid_lagged_tc, valid_fmri.T)
                # Extract correlations between lagged time course and each voxel
                # The first row (after the first element) contains these correlations
                correlation_map[i, :] = corr_matrix[0, 1:]
        
        return correlation_map

    def _validate(self):
        """
        Validate parameters

        Raises
        ------
        ParameterInputError
            if the parameters are not valid
        """
        # check lags are valid
        if not validate_less_than_or_equal_to_zero(self.negative_lag):
            raise ParameterInputError(
                message="Negative lag must be negative",
                parameters=["negative_lag"]
            )
        if not validate_greater_than_or_equal_to_zero(self.positive_lag):
            raise ParameterInputError(
                message="Positive lag must be positive",
                parameters=["positive_lag"]
            )
        if not validate_less_than_half_time_length(
            self.negative_lag, 
            self.time_length
        ):
            raise ParameterInputError(
                message="Negative lag must be less than half the time length",
                parameters=["negative_lag"]
            )
        if not validate_less_than_half_time_length(
            self.positive_lag, 
            self.time_length
        ):
            raise ParameterInputError(
                message="Positive lag must be less than half the time length",
                parameters=["positive_lag"]
            )






"""
File upload validation checks
"""
import os

from typing import List, Iterable, Tuple, Optional, Literal

import nibabel as nib
import numpy as np

from nibabel.cifti2.cifti2_axes import BrainModelAxis

from findviz.viz.io import utils


def validate_cii_brainmodel_axis(cifti_img: nib.Cifti2Image) -> bool:
    """
    Validate that the cifti image has a brainmodel axis
    """
    axes = [cifti_img.header.get_axis(i) for i in range(cifti_img.ndim)]
    return any(isinstance(axis, BrainModelAxis) for axis in axes)
    

def validate_cii_dtseries_ext(fp: str) -> bool:
    """
    validate uploaded cifti dtseries file extension (dtseries.nii)
    """
    base, ext = os.path.splitext(fp)
    if ext == '.nii':
        base, ext = os.path.splitext(base)
        if ext == '.dtseries':
            return True
        else:
            return False
    else:
        return False


def validate_cii_file_inputs(
    dtseries: Optional[str] = None,
    left_mesh: Optional[str] = None,
    right_mesh: Optional[str] = None
) -> Tuple[str, bool]:
    """
    validate uploaded cifti file inputs
    """
    msg = ''
    checks_pass = True
    
    # check if dtseries file is valid
    if dtseries is None:
        msg = 'A dtseries.nii file must be provided'
        checks_pass = False
    
    # check if left mesh file is valid
    if (left_mesh is None) and (right_mesh is None):
        msg = 'A left or right hemisphere mesh file must be provided'
        checks_pass = False
    
    return msg, checks_pass


def validate_cii_hemisphere(
    brain_model_axis: BrainModelAxis,
    hemisphere: Literal['left', 'right'],
    axis_key: str
) -> bool:
    """
    Validate that the cifti image has a brainmodel axis for the left or right hemisphere
    """
    for axis in brain_model_axis.iter_structures():
        if axis[0] == axis_key:
            return True
    return False


def validate_gii_func_ext(fp: str) -> bool:
    """
    validate uploaded gifti functional file extension (func.gii)
    """
    base, ext = os.path.splitext(fp)
    if ext == '.gii':
        base, ext = os.path.splitext(base)
        if ext == '.func':
            return True
        else:
            return False
    else:
        return False


def validate_gii_mesh_ext(fp: str) -> bool:
    """
    validate uploaded gifti mesh file extension (surf.gii)
    """
    base, ext = os.path.splitext(fp)
    if ext == '.gii':
        base, ext = os.path.splitext(base)
        if ext == '.surf':
            return True
        else:
            return False
    else:
        return False

def validate_gii_file_inputs(
    left_mesh, 
    right_mesh,
    left_func, 
    right_func
) -> Tuple[str, List[str], bool]:
    """
    Take left and right hemisphere mesh (surf.gii) and functional (func.gii)
    files and ensure the appropriate file inputs
    return a tuple of (error message, missing files, checks pass)

    """
    # initialize error message and pass state
    msg = ''
    missing = []
    checks_pass = True
    if (left_func is None) and (right_func is None):
        msg =  'A left or right hemisphere functional file must be provided'
        missing = ['left_func', 'right_func']
        checks_pass = False

    if (left_mesh is None) and (right_mesh is None):
        msg = 'A left or right hemisphere mesh file must be provided'
        missing = ['left_mesh', 'right_mesh']
        checks_pass = False

    if (right_func is None) and (right_mesh is not None):
        msg = 'A right hemisphere func file must be provided with mesh input'
        missing = ['right_func']
        checks_pass = False
        
    if (right_func is not None) and (right_mesh is None):
        msg = 'A right hemisphere mesh file must be provided with func input'
        missing = ['right_mesh']
        checks_pass = False

    if (left_func is None) and (left_mesh is not None):
        msg = 'A left hemisphere func file must be provided with mesh input'
        missing = ['left_func']
        checks_pass = False

    if (left_func is not None) and (left_mesh is None):
        msg = 'A left hemisphere mesh file must be provided with func input'
        missing = ['left_mesh']
        checks_pass = False
    
    return msg, missing, checks_pass


def validate_gii_func(gii: nib.GiftiImage) -> bool:
    """
    Validate that gifti functional file (func.gii) has a 1d-array per 
    timepoint (data array).
    """
    for darray in gii.darrays:
        if len(np.squeeze(darray.data).shape) > 1:
            return False
    return True

def validate_gii_func_mesh_len(
    gii_func: nib.GiftiImage, 
    gii_mesh: nib.GiftiImage
) -> bool:
    """
    Validate that gifti functional file (func.gii) has the same number of 
    time courses as the number of vertices in the mesh file (surf.gii).
    """
    return len(gii_func.darrays[0].data) == len(gii_mesh.darrays[0].data)

def validate_gii_mesh(gii: nib.GiftiImage) -> bool:
    """
    The gifti mesh (geometry) file should only contain two data arrays,
    coresponding to coordinates and faces arrays .
    """
    return len(gii.darrays) == 2


def validate_gii_func_len(
    gii_lh: nib.GiftiImage, 
    gii_rh: nib.GiftiImage
) -> bool:
    """
    validate that gifti functionals from left and right hemisphere
    are equal length
    """
    return len(gii_lh.darrays) == len(gii_rh.darrays)


def validate_nii_ext(fp: str) -> bool:
    """
    validate uploaded nifti file extension (either .nii or .nii.gz)
    """
    # get file extension
    ext = utils.parse_nifti_file_ext(fp)
    # if not nifti file, raise error
    if ext is None:
        return False
    return True


def validate_nii_4d(nii: nib.Nifti1Image) -> bool:
    """
    Validate that nii is 4-dimensional. This is required for functional files.
    """
    if len(nii.shape) != 4:
        return False
    else:
        return True


def validate_nii_3d(nii: nib.Nifti1Image) -> bool:
    """
    Validate nii file is 3-dimensional (not 4). This is required for anatomical
    and mask files.
    """
    if len(nii.shape) != 3:
        return False
    else:
        return True


def validate_nii_same_dim_len(nii: nib.Nifti1Image, nii_ref: nib.Nifti1Image) -> bool:
    """
    Ensure the first 3 voxel-dimensions of nifti file (nii) are the same 
    length as a reference nifti file (nii_ref)
    """
    return np.array_equal(nii.shape[:3], nii_ref.shape[:3])


def validate_nii_brain_mask(nii_mask: nib.Nifti1Image) -> bool:
    """
    Ensure that the nifti brain mask only contains 0 and 1s
    """
    # Get the data array
    data_mask = nii_mask.get_fdata()
    # Check unique values in the array only have 0s and 1s
    unique_values = np.unique(data_mask)
    return np.array_equal(unique_values, [0, 1])
    

def validate_task_ext(fp: str) -> bool:
    """
    validate task design file extension (.csv or .tsv)
    """
    base, ext = os.path.splitext(fp)
    if ext == '.csv':
        return True
    elif ext == '.tsv':
        return True
    else:
        return False


def validate_task_header_required_cols(
    header: List[str],
    required_cols: List[str]
) -> bool:
    """
    validate task design file header has necessary columns.
    It must have two columns: 'onset' or 'duration'.
    It could also have a 'task_trial' column. Case-insensitive.
    """
    # lower case all header columns
    header = [col.strip().lower() for col in header]
    missing_cols = [col for col in required_cols 
                    if col.strip().lower() not in header]
    if len(missing_cols) > 0:
        return False

    return True


def validate_task_header_duplicates(
    header: List[str],
    required_cols: List[str]
) -> bool:
    """
    Ensure there are not duplicates of required columns
    """
    # Check for duplicates only among required columns
    duplicate_cols = [
        col for col in required_cols if header.count(col) > 1
    ]
    if len(duplicate_cols) > 0:
        return False

    return True


def validate_task_tr(tr: float) -> bool:
    """
    validate TR (repetition time) is not negative or 0
    """
    if tr <= 0:
        return False
    return True


def validate_task_slicetime(slicetime: float) -> bool:
    """
    validate slicetime reference value is between 0 and 1
    """
    if slicetime > 1:
        return False
    elif slicetime < 0:
        return False
    return True


def validate_ts_task_length(ts_list: Iterable[List]) -> bool:
    """
    Take a time course list and validate there is at least one row in the
    ts or task design file.
    """
    row_count = len(ts_list)
    if row_count < 1:
        return False
    else:
        return True


def validate_ts_ext(fp: str) -> bool:
    """
    validate time course file extension (.csv or .txt)
    """
    base, ext = os.path.splitext(fp)
    if ext == '.csv':
        return True
    elif ext == '.txt':
        return True
    else:
        return False


def validate_ts_single_col(row: List) -> bool:
    """
    validate that time course has a single column
    """
    if len(row) > 1:
        return False
    return True


def validate_ts_numeric(element) -> bool:
    """
    validate that time course element is numeric
    """
    # try convert to number, if exception, return failed validation
    try:
        float(element)
    except ValueError:
        return False
    return True

def validate_ts_fmri_length(
        fmri_len: int,
        ts: np.ndarray
) -> bool:
    """
    Validate that the length of the fmri time course length matches 
    that the length of the time course. 
    """
    
    return len(ts) == fmri_len












import argparse
import os
import socket
import sys
import webbrowser

from threading import Timer

from findviz import create_app
from findviz.logger_config import setup_logger
from findviz.routes.shared import data_manager
from findviz.viz.io import gifti
from findviz.viz.io import nifti
from findviz.viz.io import cifti
from findviz.viz import exception
from findviz.viz.io.cache import Cache
from findviz.viz.io.upload import FileUpload

logger = setup_logger(__name__)

def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description='FIND Viewer')

    
    # NIFTI inputs
    nifti_group = parser.add_argument_group('NIFTI inputs', 'NIFTI file inputs')
    nifti_group.add_argument('--nifti-func', help='Functional NIFTI (nii/nii.gz)')
    nifti_group.add_argument('--nifti-anat', help='Anatomical NIFTI (nii/nii.gz)')
    nifti_group.add_argument('--nifti-mask', help='Brain mask NIFTI (nii/nii.gz)')
    
    # GIFTI inputs
    gifti_group = parser.add_argument_group('GIFTI inputs', 'GIFTI file inputs')
    gifti_group.add_argument('--gifti-left-func', help='Left hemisphere functional GIFTI (func.gii)')
    gifti_group.add_argument('--gifti-right-func', help='Right hemisphere functional GIFTI (func.gii)')
    gifti_group.add_argument('--gifti-left-mesh', help='Left hemisphere mesh GIFTI (surf.gii)')
    gifti_group.add_argument('--gifti-right-mesh', help='Right hemisphere mesh GIFTI (surf.gii)')

    # CIFTI inputs
    cifti_group = parser.add_argument_group('CIFTI inputs', 'CIFTI file inputs')
    cifti_group.add_argument('--cifti-dtseries', help='Functional CIFTI (dtseries.nii)')
    cifti_group.add_argument('--cifti-left-mesh', help='Left hemisphere mesh GIFTI (surf.gii)')
    cifti_group.add_argument('--cifti-right-mesh', help='Right hemisphere mesh GIFTI (surf.gii)')
    
    # Optional inputs
    parser.add_argument('--timeseries', nargs='+', help='Time series files')
    parser.add_argument('--ts-labels', nargs='+', help='Labels for time series files')
    parser.add_argument('--ts-headers', nargs='+', 
                        help='Whether time series files have headers (true/false)')
    
    parser.add_argument('--task-design', help='Task design file')
    parser.add_argument('--tr', type=float, help='TR value')
    parser.add_argument('--slicetime-ref', type=float, 
                       help='Slice timing reference (0-1)', default=0.5)
    parser.add_argument('--port', type=int, help='Port number', default=None)
    
    args = parser.parse_args()

    # raise FileInputError if both gifti and nifti files are present
    nifti_input = any([args.nifti_func, args.nifti_anat, args.nifti_mask])
    gifti_input = any([
        args.gifti_left_func, args.gifti_right_func, 
        args.gifti_left_mesh, args.gifti_right_mesh
    ])
    cifti_input = any([
        args.cifti_dtseries, args.cifti_left_mesh, args.cifti_right_mesh
    ])
    # Check if inputs from multiple groups are present
    if sum([nifti_input, gifti_input, cifti_input]) > 1:
        raise exception.FileInputError(
            "Only one file type (NIFTI, GIFTI, or CIFTI) can be used at a time. "
            "Please upload only one file type.",
            file_type=exception.ExceptionFileTypes.NIFTI_GIFTI_CIFTI.value,
            method='cli'
        )
    
    # Set the appropriate group flag based on which arguments are present
    if nifti_input:
        args.file_type = 'nifti'
    elif gifti_input:
        args.file_type = 'gifti'
    elif cifti_input:
        args.file_type = 'cifti'
    return args


def process_cli_inputs(args) -> None:
    """Process and validate CLI inputs using existing validation logic."""
    logger.info("Processing CLI inputs")
    # Determine file type and create FileUpload instance
    if args.file_type == 'nifti':
        fmri_type = 'nifti'
        logger.info("Nifti file type detected")
        fmri_files = {
            nifti.NiftiFiles.FUNC.value: args.nifti_func,
            nifti.NiftiFiles.ANAT.value: args.nifti_anat,
            nifti.NiftiFiles.MASK.value: args.nifti_mask
        }
    elif args.file_type == 'gifti':
        fmri_type = 'gifti'
        logger.info("Gifti file type detected")
        fmri_files = {
            gifti.GiftiFiles.LEFT_FUNC.value: args.gifti_left_func,
            gifti.GiftiFiles.RIGHT_FUNC.value: args.gifti_right_func,
            gifti.GiftiFiles.LEFT_MESH.value: args.gifti_left_mesh,
            gifti.GiftiFiles.RIGHT_MESH.value: args.gifti_right_mesh
        }
    elif args.file_type == 'cifti':
        fmri_type = 'cifti'
        logger.info("CIFTI file type detected")
        fmri_files = {
            cifti.CiftiFiles.DTSERIES.value: args.cifti_dtseries,
            cifti.CiftiFiles.LEFT_MESH.value: args.cifti_left_mesh,
            cifti.CiftiFiles.RIGHT_MESH.value: args.cifti_right_mesh
        }
    
    # Validate that all files exist
    validate_files(fmri_files)
    logger.info("FMRI files validated successfully")

    # Create additional files dict for validation
    additional_files = {}
    if args.timeseries:
        for i, ts_file in enumerate(args.timeseries):
            additional_files[f'timeseries_{i}'] = ts_file
    if args.task_design:
        additional_files['task_design'] = args.task_design
    
    # Validate additional files
    validate_files(additional_files)
    logger.info("Additional files validated successfully")

    # Create FileUpload instance
    file_upload = FileUpload(
        fmri_type,
        ts_status=bool(args.timeseries),
        task_status=bool(args.task_design),
        method='cli'
    )
    logger.info("FileUpload instance initialized")

    # Process files using existing validation logic
    uploads = file_upload.upload(
        fmri_files=fmri_files,
        ts_files=args.timeseries,
        ts_labels=args.ts_labels,
        ts_headers=args.ts_headers,
        task_file=args.task_design,
        tr=args.tr,
        slicetime_ref=args.slicetime_ref
    )
    logger.info("File uploads processed successfully")
    # pass fmri data to data manager and get viewer data
    if fmri_type == 'nifti':
        data_manager.ctx.create_nifti_state(
            func_img = uploads['nifti'][nifti.NiftiFiles.FUNC.value],
            anat_img = uploads['nifti'][nifti.NiftiFiles.ANAT.value],
            mask_img = uploads['nifti'][nifti.NiftiFiles.MASK.value]
        )
        logger.info("Nifti data manager state created successfully")
    else:
        # cifti or gifti file inputs
        data_manager.ctx.create_gifti_state(
            left_func_img=uploads['gifti'][gifti.GiftiFiles.LEFT_FUNC.value],
            right_func_img=uploads['gifti'][gifti.GiftiFiles.RIGHT_FUNC.value],
            left_mesh=uploads['gifti'][gifti.GiftiFiles.LEFT_MESH.value],
            right_mesh=uploads['gifti'][gifti.GiftiFiles.RIGHT_MESH.value]
        )
        logger.info("Gifti data manager state created successfully")

    # if timecourse data, add to viewer data
    if file_upload.ts_status:
        data_manager.ctx.add_timeseries(uploads['ts'])
        logger.info("Time series data added to viewer data")

    # if task data, add to viewer data
    if file_upload.task_status:
        data_manager.ctx.add_task_design(uploads['task'])
        logger.info("Task design data added to viewer data")

    # get viewer metadata
    viewer_metadata = data_manager.ctx.get_viewer_metadata()
    logger.info("Viewer metadata retrieved successfully")

    # create cache instance
    cache = Cache()
    # save cache
    try:
        cache.save(viewer_metadata)
    except Exception as e:
        logger.error(f"Error saving cache: {e}")

def main():
    args = parse_args()
    # If arguments were provided (besides port number), process them
    # This doesn't seem robust; might need refactoring
    # If only port number provided, skip input processing
    if len(sys.argv) == 2 and args.port:
        pass
    elif len(sys.argv) > 1:
        try:
            process_cli_inputs(args)
        except Exception as e:
            print(f"Error processing inputs: {str(e)}")
            return

    # create app
    app = create_app(clear_cache=False)
    # find free port
    port = args.port if args.port else find_free_port()
    # open browser (wait 1 second to ensure server is running)
    Timer(1, open_browser, args=(port,)).start()
    # run app
    app.run(debug=False, port=port)


def find_free_port():
    """Find an available port on the system."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(('', 0))  # Bind to any available port
        return s.getsockname()[1]  # Return the port number


def open_browser(port):
    """Open the web browser to the Flask app."""
    webbrowser.open_new(f"http://127.0.0.1:{port}")


def validate_files(files_dict: dict):
    """Validate that all provided files exist.
    
    Arguments:
    ----------
    files_dict (dict): Dictionary of file paths to check
        
    Raises:
    -------
        FileNotFoundError: If any specified file doesn't exist
    """
    missing_files = []
    for name, filepath in files_dict.items():
        if filepath and not os.path.exists(filepath):
            missing_files.append(filepath)
    
    if missing_files:
        raise FileNotFoundError(
            f"The following files were not found: {', '.join(missing_files)}"
        )


"""
Log retrieval routes
"""
import os
import re
import glob
from datetime import datetime, timedelta
from typing import List, Dict

from flask import Blueprint, request, jsonify

from findviz.logger_config import setup_logger
from findviz.routes.utils import handle_route_errors, Routes

# Set up a logger for the app
logger = setup_logger(__name__)

logs_bp = Blueprint('logs', __name__)

@logs_bp.route(Routes.GET_LOG_ENTRIES.value, methods=['GET'])
@handle_route_errors(
    error_msg='Error retrieving log entries',
    log_msg='Log entries request successful',
    route=Routes.GET_LOG_ENTRIES
)
def get_log_entries():
    """Get recent log entries"""
    # Get parameters with defaults
    max_entries = int(request.args.get('max_entries', 100))
    since_minutes = int(request.args.get('since_minutes', 15))
    log_file = request.args.get('log_file', None)  # Allow specifying which log file
    
    # Limit max_entries to a reasonable value to prevent performance issues
    max_entries = min(max_entries, 1000)
    
    # If no log file specified, find the most recent one
    if log_file is None:
        log_file = find_most_recent_log_file()
    
    # Get log entries
    log_entries = get_recent_log_entries(
        max_entries=max_entries,
        since_minutes=since_minutes,
        log_file_path=os.path.join('logs', log_file)
    )
    
    return jsonify(log_entries)

@logs_bp.route(Routes.GET_LOG_FILES.value, methods=['GET'])
@handle_route_errors(
    error_msg='Error retrieving log files',
    log_msg='Log files request successful',
    route=Routes.GET_LOG_FILES
)
def get_log_files():
    """Get available log files"""
    log_dir = os.path.join(os.getcwd(), 'logs')
    
    if not os.path.exists(log_dir):
        return jsonify([])
    
    # Get all log files and their modification times
    log_files = []
    
    # Get all run-specific log files (app-run-*.log)
    run_log_files = glob.glob(os.path.join(log_dir, 'app-run-*.log'))
    
    for file_path in run_log_files:
        file_name = os.path.basename(file_path)
        mod_time = os.path.getmtime(file_path)
        size = os.path.getsize(file_path)
        
        # Extract timestamp from filename
        timestamp_match = re.search(r'app-run-(\d{8}-\d{6})\.log', file_name)
        timestamp_str = None
        if timestamp_match:
            # Convert YYYYMMDD-HHMMSS to a readable format
            ts = timestamp_match.group(1)
            try:
                dt = datetime.strptime(ts, "%Y%m%d-%H%M%S")
                timestamp_str = dt.strftime("%Y-%m-%d %H:%M:%S")
            except ValueError:
                timestamp_str = None
        
        log_files.append({
            'name': file_name,
            'modified': datetime.fromtimestamp(mod_time).isoformat(),
            'size': size,
            'timestamp': timestamp_str or "Unknown"
        })
    
    # Sort by modification time (newest first)
    log_files.sort(key=lambda x: x['modified'], reverse=True)
    
    return jsonify(log_files)


def find_most_recent_log_file() -> str:
    """
    Find the most recent run log file in the logs directory.
    
    Returns
    -------
    str
        The name of the most recent log file, or None if none found
    """
    log_dir = os.path.join(os.getcwd(), 'logs')
    
    if not os.path.exists(log_dir):
        logger.warning(f"Log directory not found: {log_dir}")
        return None
    
    # Look for all run-specific log files
    run_log_files = glob.glob(os.path.join(log_dir, 'app-run-*.log'))
    
    if not run_log_files:
        logger.warning("No run log files found")
        return None
    
    # Find the most recently modified file
    most_recent = max(run_log_files, key=os.path.getmtime)
    return os.path.basename(most_recent)


def get_recent_log_entries(
    max_entries: int = 100, 
    since_minutes: int = 15, 
    log_file_path: str = None
) -> List[Dict]:
    """
    Retrieve recent log entries from the log file.
    
    Parameters
    ----------
    max_entries : int
        Maximum number of log entries to return
    since_minutes : int
        Only return entries from the last N minutes
    log_file_path : str
        Path to the log file
    
    Returns
    -------
    List[Dict]
        List of log entries with timestamp, level, source, and message
    """
    entries = []
    
    # Check if log file exists
    if log_file_path is None or not os.path.exists(log_file_path):
        return [{"timestamp": datetime.now().isoformat(), 
                "level": "WARNING", 
                "source": "log_utils", 
                "message": f"Log file not found: {log_file_path}"}]
    
    # Calculate cutoff time
    cutoff_time = datetime.now() - timedelta(minutes=since_minutes)
    
    try:
        with open(log_file_path, 'r') as log_file:
            # Read the file from the end, which is more efficient for large logs
            lines = log_file.readlines()
            
            # Process the most recent lines first (reversed)
            for line in reversed(lines):
                # Skip empty lines
                if not line.strip():
                    continue
                
                # Parse log entry with regex
                # Format: 2023-05-21 14:30:45,123 - module_name - LEVEL - Message
                match = re.match(
                    r'(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2},\d{3}) - ([\w\._]+) - (\w+) - (.*)', 
                    line
                )

                if match:
                    timestamp_str, source, level, message = match.groups()
                    
                    try:
                        # Parse timestamp - handle milliseconds correctly
                        timestamp_str = timestamp_str.replace(',', '.')
                        timestamp = datetime.strptime(timestamp_str, '%Y-%m-%d %H:%M:%S.%f')
                        
                        # For run-specific logs, we might want to show all entries
                        # regardless of time, since each run is a discrete session
                        # But we'll keep the time filter as an option
                        if since_minutes > 0 and timestamp < cutoff_time:
                            continue
                        
                        entries.append({
                            "timestamp": timestamp.isoformat(),
                            "level": level,
                            "source": source,
                            "message": message.strip()
                        })
                        
                        # Stop if we have enough entries
                        if len(entries) >= max_entries:
                            break
                    except ValueError as e:
                        # Log the error but continue processing
                        logger.warning(f"Error parsing timestamp '{timestamp_str}': {e}")
                        continue
                else:
                    # For multiline log entries, append to the last entry's message
                    if entries:
                        entries[0]["message"] += "\n" + line.strip()
        
        # Return entries in chronological order (oldest first)
        return list(reversed(entries))
    
    except Exception as e:
        return [{"timestamp": datetime.now().isoformat(), 
                "level": "ERROR", 
                "source": "log_utils", 
                "message": f"Error reading log file: {str(e)}"}]


"""
Main application routes.
"""
from flask import Blueprint, render_template

from findviz.routes.shared import data_manager

main_bp = Blueprint('main', __name__)

@main_bp.route('/')
def index():
    """Serve the main application page."""
    return render_template('index.html')

# display analysis view for a specific analysis route
@main_bp.route('/analysis_view/<analysis>')
def analysis_view(analysis):
    """Display the analysis view for a specific analysis."""
    if analysis == 'average':
        data_manager.switch_context('average')
    elif analysis == 'correlate':
        data_manager.switch_context('correlate')
    
    # get the plot type
    plot_type = data_manager.ctx.fmri_file_type

    return render_template(
        'analysis.html', 
        plot_type=plot_type,
        analysis=analysis
    )


"""
Utilities for handling time course and task design uploads
"""
from enum import Enum
from pathlib import Path
from typing import Dict, List, Literal, Optional, Union, TypedDict, Any, Tuple
from werkzeug.datastructures import FileStorage

import numpy as np
import numpy.typing as npt

from flask import request
from nilearn.glm.first_level import compute_regressor

from findviz.viz import exception
from findviz.viz.io import validate
from findviz.viz.io import utils


# Type aliases
FilePath = Union[str, Path]
FileInput = Union[FilePath, FileStorage]


# task design output fields
class ConditionDict(TypedDict):
    block: List[float]
    hrf: List[float]

class TaskDesignDict(TypedDict):
    task_regressors: Dict[str, ConditionDict]
    tr: float
    slicetime_ref: float

# time course output fields
class TimeCourseDict(TypedDict):
    ts_label: List[float]


# Expected time course file upload inputs
class TimeCourseFiles(Enum):
    FILES = 'ts_files'
    LABELS = 'ts_labels'
    HEADERS = 'ts_headers'


# Time course fields for single time course upload
class SingleTimeCourseFiles(Enum):
    FILE = 'ts_file'
    LABEL = 'ts_label'
    HEADER = 'ts_header'


# Expected task design file upload inputs
class TaskDesignFiles(Enum):
    FILE = 'task_file'
    TR = 'tr'
    SLICETIME = 'slicetime_ref'


# Task Design Output Fields
class TaskDesignFields(Enum):
    ONSET = 'onset'
    DURATION = 'duration'
    TRIAL_TYPE = 'trial_type'


# timecourse/task form fields in upload modal
browser_fields = {
    TimeCourseFiles.FILES.value: 'time-series-file',
    TimeCourseFiles.LABELS.value: 'time-series-label',
    TimeCourseFiles.HEADERS.value: 'has-header',
    TaskDesignFiles.FILE.value: 'task-design-file',
    TaskDesignFiles.TR.value: 'task-design-tr',
    TaskDesignFiles.SLICETIME.value: 'task-design-slicetime'
}

class TaskDesignUpload:
    """
    Class for handling task design file uploads.

    Attributes:
        method (str): The method used for file uploads (cli , browser)
        default_trial_label (str): The default trial label for task design files
        that did not have 'trial_type' field
    """

    def __init__(
        self,
        method: Literal['cli', 'browser'],
        default_trial_label: str = 'task'
    ):
        self.method = method
        self.default_trial_label = default_trial_label

    def upload(
        self,
        fmri_len: int,
        task_file: Optional[Union[str, Path]] = None,
        tr: Optional[float] = None,
        slicetime_ref: Optional[float] = None
    ) -> TaskDesignDict:
        """
        Get task design files and fields uploaded from cli or browser.

        Parameters
        ----------
        fmri_len : int
            Length of functional MRI time course
        task_file : Union[str, Path], optional
            Path to task design file
        tr : float, optional
            Repetition time value
        slicetime_ref : float, optional
            Slice timing reference value (0-1)

        Returns
        -------
        TaskDesignDict
            Dictionary containing task design data and parameters

        Raises
        ------
        FileInputError
            If required files are missing or invalid
        FileValidationError
            If file contents fail validation
        FileUploadError
            If there are issues reading the files
        """
        try:
            if self.method == 'cli':
                file_uploads = {
                    TaskDesignFiles.FILE.value: task_file,
                    TaskDesignFiles.TR.value: tr,
                    TaskDesignFiles.SLICETIME.value: slicetime_ref
                }
            elif self.method == 'browser':
                file_uploads = self._get_browser_input()
        except exception.FileInputError as e:
            # propagate file input returned from get files function
            raise e

        task_tr = file_uploads[TaskDesignFiles.TR.value]
        task_slicetime = file_uploads[TaskDesignFiles.SLICETIME.value]

        # first, validate tr and slicetime values are provided
        if task_tr == '' or task_tr is None:
            raise exception.FileInputError(
                'TR is required for task design file',
                exception.ExceptionFileTypes.TASK.value,
                self.method,
                [browser_fields[TaskDesignFiles.TR.value]]
            )
        if task_slicetime == '' or task_slicetime is None:
            raise exception.FileInputError(
                'Slicetime reference is required for task design file',
                exception.ExceptionFileTypes.TASK.value,
                self.method,
                [browser_fields[TaskDesignFiles.SLICETIME.value]]
            )

        # validate tr and slicetime values are numeric
        if not validate.validate_ts_numeric(task_tr):
            raise exception.FileValidationError(
                'Provided TR for task design file is not numeric: '
                f'{task_tr}. Please check entry',
                validate.validate_ts_numeric.__name__,
                exception.ExceptionFileTypes.TASK.value,
                [browser_fields[TaskDesignFiles.FILE.value]]
            )

        if not validate.validate_ts_numeric(task_slicetime):
            raise exception.FileValidationError(
                f'Provided slicetime reference for task design file is not '
                f'numeric: {task_slicetime}. Please check entry',
                validate.validate_ts_numeric.__name__,
                exception.ExceptionFileTypes.TASK.value,
                [browser_fields[TaskDesignFiles.FILE.value]]
            )

        task_tr = float(task_tr)
        task_slicetime = float(task_slicetime)
        # validate tr is not negative
        if not validate.validate_task_tr(task_tr):
            raise exception.FileValidationError(
                f'TR must not be less than zero: {task_tr}. '
                'Please check entry',
                validate.validate_task_tr.__name__,
                exception.ExceptionFileTypes.TASK.value,
                [browser_fields[TaskDesignFiles.FILE.value]]
            )

        # validate slicetime is b/w 0 and 1
        if not validate.validate_task_slicetime(task_slicetime):
            raise exception.FileValidationError(
                f'Slicetime reference must between 0 and 1: {task_slicetime}. '
                'Please check entry',
                validate.validate_task_slicetime.__name__,
                exception.ExceptionFileTypes.TASK.value,
                [browser_fields[TaskDesignFiles.FILE.value]]
            )

        # read task design file
        try:
            task_data = read_task_file(
                file_uploads[TaskDesignFiles.FILE.value],
                self.default_trial_label,
                self.method
            )
        # raise exception to be handled higher in stack
        except Exception as e:
            raise e

        # get task regressors
        task_reg = get_task_regressors(
            task_data, 
            task_tr, 
            task_slicetime, 
            fmri_len
        )
        task_out = {
            TaskDesignFiles.SLICETIME.value: task_slicetime,
            TaskDesignFiles.TR.value: task_tr,
            'task_regressors': task_reg
        }

        return task_out

    def _get_browser_input(self) -> Dict[str, Any]:
        """
        Get task design inputs from browser request.

        Returns
        -------
        Dict[str, Any]
            Dictionary containing file and parameter inputs

        Raises
        ------
        FileInputError
            If file extension is invalid
        """
        task_file = request.files.get(TaskDesignFiles.FILE.value)
        task_tr = request.form.get(TaskDesignFiles.TR.value)
        task_slicetime = request.form.get(TaskDesignFiles.SLICETIME.value)

        file_upload = {
            TaskDesignFiles.FILE.value: task_file,
            TaskDesignFiles.TR.value: task_tr,
            TaskDesignFiles.SLICETIME.value: task_slicetime
        }
        # check file extension
        if not validate.validate_task_ext(
            utils.get_filename(task_file.filename)
        ):
            raise exception.FileInputError(
                f'Unrecognized file extension for {task_file.filename}. '
                'Only .csv or .tsv are allowed.',
                exception.ExceptionFileTypes.TASK.value, 
                self.method,
                [browser_fields[TaskDesignFiles.FILE.value]]
            )

        return file_upload


class TimeCourseUpload:
    """
    Class for handling time course file uploads.

    Attributes:
        method (str): The method used for file uploads (cli , browser)
    """

    def __init__(
        self,
        method: Literal['cli', 'browser']
    ):
        self.method = method

    def upload(
        self, 
        fmri_len: int, 
        ts_files: Optional[List[Union[str, Path]]] = None,
        ts_labels: Optional[List[str]] = None,
        ts_headers: Optional[List[bool]] = None,
    ) -> TimeCourseDict:
        """
        Get time course files and fields uploaded from cli or browser.

        Parameters
        ----------
        fmri_len : int
            Length of functional MRI time course
        ts_files : List[Union[str, Path]], optional
            List of paths to time series files
        ts_labels : List[str], optional
            List of labels for time series files
        ts_headers : List[bool], optional
            List of boolean flags indicating if time series files have headers

        Returns
        -------
        List[TimeCourseDict]
            List of dictionaries containing time course arrays and labels

        Raises
        ------
        FileInputError
            If files are missing or invalid
        FileValidationError
            If file contents fail validation
        """
        try:
            if self.method == 'cli':
                file_uploads = []
                for file, label, header in zip(ts_files, ts_labels, ts_headers):
                    single_upload = {
                        SingleTimeCourseFiles.FILE.value: file,
                        SingleTimeCourseFiles.LABEL.value: label,
                        SingleTimeCourseFiles.HEADER.value: header
                    }
                    file_uploads.append(single_upload)
            elif self.method == 'browser':
                file_uploads = self._get_browser_input()
        except exception.FileInputError as e:
            # propagate file input returned from get files function
            raise e

        # check for duplicate time series labels in input
        ts_labels = [
            file[SingleTimeCourseFiles.LABEL.value]
            for file in file_uploads
        ]
        dups = self._check_duplicate_labels(ts_labels)
        if len(dups) > 0:
            # get indices of duplicate labels
            dups_indx = [indx for indx, label in enumerate(ts_labels)
                          if label == dups[0]]
            raise exception.FileInputError(
                f'Duplicate time course labels found: {dups}. '
                'Please use unique labels.',
                'task', self.method,
                [browser_fields[TimeCourseFiles.FILES.value]],
                index=dups_indx
            )

        # Loop through timecourses and read
        ts_out = {}
        for i, file_package in enumerate(file_uploads):
            ts_file = file_package[SingleTimeCourseFiles.FILE.value]
            ts_label = file_package[SingleTimeCourseFiles.LABEL.value]
            ts_header = file_package[SingleTimeCourseFiles.HEADER.value]
            # check file extension
            if not validate.validate_ts_ext(
                utils.get_filename(ts_file.filename)
            ):
                raise exception.FileInputError(
                    'Unrecognized file extension for '
                    f'{utils.get_filename(ts_file)}.'
                     ' Only .csv or .txt are allowed.',
                    exception.ExceptionFileTypes.TIMECOURSE.value, self.method,
                    [browser_fields[TimeCourseFiles.FILES.value]],
                    index=[i]
                )
            # read time course file
            try:
                ts_array = read_ts_file(ts_file, ts_header, self.method, index=i)
            # raise exception to be handled higher in stack
            except Exception as e:
                raise e
                
            # check time course is the same length as fmri time course
            if not validate.validate_ts_fmri_length(fmri_len, ts_array):
                raise exception.FileValidationError(
                    f"length of {SingleTimeCourseFiles.FILE.value} "
                    f"({len(ts_array)}) is not the same length as "
                    f" fmri volumes ({fmri_len})",
                    validate.validate_ts_fmri_length.__name__,
                    exception.ExceptionFileTypes.TIMECOURSE.value,
                    [browser_fields[TimeCourseFiles.FILES.value]],
                    index=[i]
                )
            
            # package array in dict
            ts_out[ts_label] = ts_array

        return ts_out

    def _get_browser_input(self) -> List[Dict[str, Any]]:
        ts_file = request.files.getlist(TimeCourseFiles.FILES.value)
        ts_labels = request.form.getlist(TimeCourseFiles.LABELS.value)
        ts_headers = request.form.getlist(TimeCourseFiles.HEADERS.value)
        file_uploads = []
        for file, label, header in zip(ts_file, ts_labels, ts_headers):
            # convert header form data to boolean
            header_bool = True if header == 'true' else False
            single_upload = {
                SingleTimeCourseFiles.FILE.value: file,
                SingleTimeCourseFiles.LABEL.value: label,
                SingleTimeCourseFiles.HEADER.value: header_bool
            }
            file_uploads.append(single_upload)

        return file_uploads

    def _check_duplicate_labels(self, labels: List[str]) -> List[str]:
        """
        Check for duplicate time series labels.

        Parameters
        ----------
        labels : List[str]
            List of time series labels

        Returns
        -------
        List[str]
            List of duplicate labels found
        """
        duplicates = []
        seen = set()

        for element in labels:
            if element in seen:
                duplicates.append(element)
            else:
                seen.add(element)

        return duplicates


def get_ts_header(
    file: FileInput,
    file_index: int
) -> str:
    """
    Get first row of time course file as header and validate time course file.

    Parameters
    ----------
    file : FileInput
        Time course file from browser or CLI
    file_index : int
        Index of the time course file in multi-file upload

    Returns
    -------
    str
        The first row (representing the header)

    Raises
    ------
    FileInputError
        If file extension is invalid
    FileUploadError
        If there are issues reading the file
    """
    # load time course file and validate
    # check file extension
    filename = utils.get_filename(file)
    if not validate.validate_ts_ext(filename):
        raise exception.FileInputError(
            f'Unrecognized file extension for {filename}.'
                ' Only .csv or .txt are allowed.',
            exception.ExceptionFileTypes.TIMECOURSE.value, 'browser',
            [browser_fields[TimeCourseFiles.FILES.value]],
            index=[file_index]
        )
    # get header
    try:
        delimiter = ','
        reader = utils.get_csv_reader(file, delimiter, 'browser')
        header = next(reader)
    # raise generic exception to be handled higher in stack
    except Exception as e:
        raise exception.FileUploadError(
            f'Error in reading time course file:  {filename}',
            exception.ExceptionFileTypes.TIMECOURSE.value, 'browser',
            [browser_fields[TimeCourseFiles.FILES.value]],
            index=[file_index]
        ) from e

    # return first row
    return str(header[0])


def read_task_file(
    file: FileInput,
    default_trial_label: str,
    method: Literal['cli', 'browser']
) -> Dict[str, Any]:
    """
    Read task design file based on method of upload.

    Parameters
    ----------
    file : FileInput
        Task design file from browser or CLI
    default_trial_label : str
        Default label for trials without type
    method : Literal['cli', 'browser']
        Upload method

    Returns
    -------
    TaskDesignDict
        Dictionary containing task design data

    Raises
    ------
    ValueError
        If file object is invalid
    FileValidationError
        If file contents fail validation
    """
     # Get filename from file object
    filename = utils.get_filename(file)
    
    # get file extension (should be .tsv or .csv)
    ext = utils.get_file_ext(filename)

    if ext == '.tsv':
        delimiter = '\t'
    elif ext == '.csv':
        delimiter = ','

    # get reader for file input
    try:
        reader = utils.get_csv_reader(file, delimiter, method)
    # raise generic exception to be handled higher in stack
    except Exception as e:
        raise e

    # get and validate header
    header = next(reader)
    required_cols = [
        TaskDesignFields.ONSET.value,
        TaskDesignFields.DURATION.value
    ]
    # validate required columns ('onset' and 'duration')
    if not validate.validate_task_header_required_cols(header, required_cols):
        raise exception.FileValidationError(
            f'Task design file is missing "{required_cols[0]}" or '
            f'"{required_cols[1]}" column:  {filename}',
            validate.validate_task_header_required_cols.__name__,
            exception.ExceptionFileTypes.TASK.value,
            [browser_fields[TaskDesignFiles.FILE.value]]
        )

    # validate there are no duplicate 'onset' or 'duration' columns
    if not validate.validate_task_header_duplicates(header, required_cols):
        raise exception.FileValidationError(
            f'Task design file has multiple "{required_cols[0]}" or '
            f'"{required_cols[1]}" column:  {filename}',
            validate.validate_task_header_duplicates.__name__,
            exception.ExceptionFileTypes.TASK.value,
            [browser_fields[TaskDesignFiles.FILE.value]]
        )

    # lower case and strip all columns in header
    header = [c.strip().lower() for c in header]
    # get indices of onset, duration columns
    onset_idx = header.index(TaskDesignFields.ONSET.value)
    duration_idx = header.index(TaskDesignFields.DURATION.value)

    # check whether optional trial type column is included
    try:
        trial_type_idx = header.index(TaskDesignFields.TRIAL_TYPE.value)
        trial_type = True
    except ValueError:
        trial_type_idx = None
        trial_type = False

    # Loop through rows of task design file
    task_out = {
        TaskDesignFields.ONSET.value: [],
        TaskDesignFields.DURATION.value: [],
        TaskDesignFields.TRIAL_TYPE.value: []
    }

    for i, row in enumerate(reader):
        # Check that onset and duration column are numeric
        if not validate.validate_ts_numeric(row[onset_idx]):
            raise exception.FileValidationError(
                f'Non-numeric entry - {row[onset_idx]} - found in '
                f'onset column of task design file {filename}. All '
                'elements must be numeric.',
                validate.validate_ts_numeric.__name__,
                exception.ExceptionFileTypes.TASK.value,
                [browser_fields[TaskDesignFiles.FILE.value]]
            )

        if not validate.validate_ts_numeric(row[duration_idx]):
            raise exception.FileValidationError(
                f'Non-numeric entry - {row[duration_idx]} - found in onset '
                f'column of task design file {filename}. All elements must be numeric.',
                validate.validate_ts_numeric.__name__,
                exception.ExceptionFileTypes.TASK.value,
                [browser_fields[TaskDesignFiles.FILE.value]]
            )
        # if checks passed, append to task out dict
        task_out[TaskDesignFields.ONSET.value].append(row[onset_idx])
        task_out[TaskDesignFields.DURATION.value].append(row[duration_idx])
        # append trial type if passed, otherwise default label
        if trial_type:
            task_out[TaskDesignFields.TRIAL_TYPE.value].append(
                row[trial_type_idx]
            )
        else:
            task_out[TaskDesignFields.TRIAL_TYPE.value].append(
                default_trial_label
            )
    return task_out


def read_ts_file(
    file: FileInput,
    header: bool,
    method: Literal['cli', 'browser'],
    index: Optional[int] = None,
    validate_numeric: bool = True
) -> List[float]:
    """
    Read time course file based on method of upload.

    Parameters
    ----------
    file : FileInput
        Time course file from browser or CLI
    header : bool
        Whether file has a header row
    method : Literal['cli', 'browser']
        Upload method
    index : Optional[int]
        Index of file in multi-file upload
    validate_numeric : bool
        Whether to validate numeric content

    Returns
    -------
    List[float]
        Time course as list of floats
    """
    filename = utils.get_filename(file)

    # get file reader for file input
    delimiter = ','
    try:
        reader = utils.get_csv_reader(file, delimiter, method)
    # raise generic exception to be handled higher in stack
    except Exception as e:
        raise exception.FileUploadError(
            f'Error in reading time course file:  {filename}',
            exception.ExceptionFileTypes.TIMECOURSE.value, method,
            [browser_fields[TimeCourseFiles.FILES.value]],
            index=index
        ) from e
    
    # skip first row, if header
    if header:
        next(reader)
    
    # read timecourse from list
    ts_raw = list(reader)
    # ensure there are rows in the data
    if not validate.validate_ts_task_length(ts_raw):
        raise exception.FileValidationError(
                f'No data found in time course file {filename}. '
                'Please check file',
                validate.validate_ts_task_length.__name__,
                exception.ExceptionFileTypes.TIMECOURSE.value,
                [browser_fields[TimeCourseFiles.FILES.value]],
                index=index
            )

    # Loop through rows of time course file and validate
    ts_array = []
    for i, row in enumerate(ts_raw):
        if not validate.validate_ts_single_col(row):
            raise exception.FileValidationError(
                f'Multiple columns found in time course file {filename}. '
                 'Only one column is allowed.',
                validate.validate_ts_single_col.__name__,
                exception.ExceptionFileTypes.TIMECOURSE.value,
                [browser_fields[TimeCourseFiles.FILES.value]],
                index=index
            )
        
        if validate_numeric:
            if not validate.validate_ts_numeric(row[0]):
                if header:
                    row_i = i
                else:
                    row_i = i+1
                raise exception.FileValidationError(
                    f'Non-numeric entry found in time course file {filename} on Line {row_i}.'
                    ' All elements must be numeric. If there is a header, '
                    'specify in input.',
                    validate.validate_ts_numeric.__name__,
                    exception.ExceptionFileTypes.TIMECOURSE.value,
                    [browser_fields[TimeCourseFiles.FILES.value]],
                    index=index
                )
        # if checks passed, append float of first index of row
        ts_array.append(float(row[0]))
    
    # return list of floats
    return ts_array

def get_task_regressors(
    task_events: Dict[str, Any], 
    task_tr: float,
    task_slicetime_ref: float,
    fmri_len: int
) -> Dict[str, Dict[str, List[float]]]:
    """
    Get task design regressors from task events dataframe

    Arguments:
    ----------
        task_events: task events dict with 'onset', 'duration', 'trial_type' keys
            (optional: 'trial_type')
        task_tr: float,
            TR of task design file
        task_slicetime_ref: float,
            Slicetime reference of task design file
        fmri_len: length of fmri
    
    Returns:
    --------
        task_reg: task design regressors
    """
    # calculate frame times based on lenght of fmri, slicetime ref and tr
    frame_times =  task_tr * (np.arange(fmri_len) + task_slicetime_ref)
    # initialize task regressors dict
    task_reg = {}

    trial_type = task_events['trial_type']
    onset = task_events['onset']
    duration = task_events['duration']

    # get conditions
    conditions = list(set(trial_type))

    # Loop through each condition and create regressors
    for c in conditions:
        task_reg[c] = {}
        # get row indices of condition events
        condition_idx = [
            i for i, r in enumerate(trial_type)
            if r == c
        ]
        # Get onset of events in condition
        c_onsets = [
            float(onset[i]) for i in condition_idx
        ]
        # Get duration of events in condition
        c_duration = [
            float(duration[i]) for i in condition_idx
        ]
        # generate dummy amplitude values of 1s
        c_amp = [1 for i in condition_idx]
        # package condition data into nested list
        conditions_desc = [c_onsets, c_duration, c_amp]

        # use nilearn to compute task regression w/o convolution
        cond_reg, _ = compute_regressor(
            conditions_desc, hrf_model=None, frame_times=frame_times
        )
        task_reg[c]['block'] = cond_reg[:,0].tolist()

        # use nilearn to compute task regression w hrf convolution
        cond_reg, _ = compute_regressor(
            conditions_desc, hrf_model='glover', frame_times=frame_times
        )
        task_reg[c]['hrf'] = cond_reg[:,0].tolist()
        
    return task_reg


"""
FINDVIZ Viewer Module

This module provides functionality for managing and visualizing neuroimaging data.
It includes components for handling both NIFTI and GIFTI data formats, as well as
supporting time series and task design data visualization.

Components:
    - data_manager: Singleton class for managing visualization state
    - types: Type definitions for visualization states and data structures
    - state: Visualization state classes
    - utils: Utility functions for data processing and metadata extraction

Example:
    >>> from findviz.viz.viewer import data_manager
    >>> dm = data_manager.DataManager()
    >>> dm.create_nifti_state(func_img=nifti_img)
"""

from findviz.viz.viewer.utils import (
    package_gii_metadata,
    package_nii_metadata
)

from findviz.viz.viewer.types import (
    ViewerMetadataNiftiDict,
    ViewerMetadataGiftiDict,
    ViewerDataNiftiDict,
    ViewerDataGiftiDict
)

from findviz.viz.viewer.data_manager import DataManager
from findviz.viz.viewer.state.viz_state import (
    NiftiVisualizationState,
    GiftiVisualizationState
)
from findviz.viz.viewer.state import components

__all__ = [
    # Data Manager
    'DataManager',
    
    # states
    'components',
    'NiftiVisualizationState',
    'GiftiVisualizationState',
    
    # Type Definitions
    'ViewerMetadataNiftiDict',
    'ViewerMetadataGiftiDict',
    'ViewerDataNiftiDict',
    'ViewerDataGiftiDict',
    
    # Utility Functions
    'package_gii_metadata',
    'package_nii_metadata'
] 


"""
FINDVIZ Visualization Module

This module provides core functionality for handling, processing, and visualizing
neuroimaging data in the FINDVIZ application.

Submodules:
    - io: Input/output operations for neuroimaging data formats (NIFTI, GIFTI)
    - viewer: Components for managing and visualizing neuroimaging data
    - analysis: Tools for analyzing neuroimaging data
    - transforms: Format conversions between neuroimaging data types
    - exception: Custom exceptions for error handling

Example:
    >>> from findviz.viz.io import FileUpload
    >>> from findviz.viz.viewer import DataManager
    >>> from findviz.viz.transforms import nifti_to_array
    >>> 
    >>> # Upload and process neuroimaging data
    >>> uploader = FileUpload()
    >>> data = uploader.upload(fmri_files={'func': 'func.nii.gz'})
    >>> 
    >>> # Convert NIFTI to array for processing
    >>> array_data = nifti_to_array(data['nifti']['func'])
    >>> 
    >>> # Create visualization state
    >>> dm = DataManager()
    >>> dm.create_nifti_state(func_img=data['nifti']['func'])
"""

from findviz.viz import io
from findviz.viz import viewer
from findviz.viz import analysis
from findviz.viz import transforms
from findviz.viz import exception

# Version information
__version__ = '0.1.0'

__all__ = [
    'io',          # Input/output operations
    'viewer',      # Visualization components
    'analysis',    # Analysis tools
    'transforms',  # Format conversions
    'exception',   # Custom exceptions
]

# Module level logger
import logging
logger = logging.getLogger(__name__)


"""
Module for caching uploaded files and data
"""
import atexit
import json
import signal
import tempfile

from pathlib import Path

import nibabel as nib
import numpy as np

from findviz.logger_config import setup_logger

logger = setup_logger(__name__)
class Cache:
    """Class for managing temporary cache of uploaded files and data"""
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Cache, cls).__new__(cls)
            # Initialize instance attributes here
            cls._instance._initialize()
        return cls._instance

    def _initialize(self):
        """Initialize the cache instance"""
        self.temp_dir = Path(tempfile.gettempdir()) / "findviz_cache"
        self.temp_dir.mkdir(exist_ok=True)
        self.cache_file = self.temp_dir / "viewer_cache.json"
        
        # Register cleanup on exit
        atexit.register(self.cleanup)

        # Register signal handlers
        try:
            # Handle Ctrl+C (SIGINT)
            signal.signal(signal.SIGINT, cleanup_handler(self))
            # Handle termination (SIGTERM)
            signal.signal(signal.SIGTERM, cleanup_handler(self))
        except (ValueError, AttributeError):
            pass

    def _ensure_temp_dir(self):
        """Ensure temporary directory exists"""
        self.temp_dir.mkdir(parents=True, exist_ok=True)

    def save(self, data):
        """Save data to cache file.
        
        Parameters
        ----------
        data : dict
            Data to cache
            
        Raises
        ------
        IOError
            If saving fails
        """
        try:
            # Create parent directory if it doesn't exist
            self.cache_file.parent.mkdir(parents=True, exist_ok=True)
            # serialize data
            serialized_data = self._serialize_data(data)
            # save to cache file
            with open(self.cache_file, 'w') as f:
                json.dump(serialized_data, f)
            logger.info(f"Viewer metadata saved to cache at {self.cache_file}")
        except Exception as e:
            logger.error(f"Failed to save cache: {str(e)}")
            raise IOError(f"Failed to save cache: {str(e)}") from e

    def load(self):
        """
        Load data from cache file
        
        Returns
        -------
        dict
            Cached data if exists, None otherwise
        """
        try:
            if self.cache_file.exists():
                with open(self.cache_file, 'r') as f:
                    return json.load(f)
            return None
        except Exception as e:
            logger.error(f"Failed to load cache: {str(e)}")
            raise IOError(f"Failed to load cache: {str(e)}") from e

    def clear(self, during_shutdown=False):
        """Clear all cached data"""
        if not during_shutdown:
            logger.info("Clearing cache...")
        try:
            if self.cache_file.exists():
                self.cache_file.unlink()
                if not during_shutdown:
                    logger.info("Cache cleared successfully")
        except Exception as e:
            if not during_shutdown:
                logger.error(f"Failed to clear cache: {str(e)}")
            else:
                print(f"Warning: Failed to clear cache: {str(e)}")
            if not during_shutdown:
                raise IOError(f"Failed to clear cache: {str(e)}")

    def exists(self):
        """
        Check if cache exists
        
        Returns
        -------
        bool
            True if cache exists, False otherwise
        """
        return self.cache_file.exists()

    def get_cache_path(self):
        """
        Get path to cache file
        
        Returns
        -------
        Path
            Path to cache file
        """
        return self.cache_file
    

    def cleanup(self):
        """Clean up temporary files on exit"""
        try:
            self.clear(during_shutdown=True)  # Pass flag to avoid logging
            if self.temp_dir.exists():
                # Only remove if empty
                if not any(self.temp_dir.iterdir()):
                    self.temp_dir.rmdir()
        except Exception as e:
            print(f"Warning: Failed to clean up cache: {str(e)}")
    
    def _serialize_data(self, data):
        """Serialize data before saving to cache.
        
        Handles Nifti1Image and GiftiImage objects by converting them to 
        serializable dictionaries containing essential information.
        """
        if isinstance(data, dict):
            return {k: self._serialize_data(v) for k, v in data.items()}
        elif isinstance(data, list):
            return [self._serialize_data(item) for item in data]
        elif isinstance(data, nib.Nifti1Image):
            return {
                '_type': 'Nifti1Image',
                'shape': data.shape,
                'affine': data.affine.tolist(),
                'header': {
                    'dim': data.header['dim'].tolist(),
                    'pixdim': data.header['pixdim'].tolist()
                }
            }
        elif isinstance(data, nib.GiftiImage):
            return {
                '_type': 'GiftiImage',
                'darrays': [arr.data.shape for arr in data.darrays]
            }
        elif isinstance(data, np.ndarray):
            return data.tolist()
        return data

def cleanup_handler(cache):
    """Handler for cleanup on exit"""
    def handler(signum, frame):
        cache.clear(during_shutdown=True)  # Pass the flag here too
        # Exit gracefully
        exit(0)
    return handler



"""
Utilities for handling cifti file uploads. The left and/or right hemisphere
time series from the dtseries.nii file is converted to gifti files 
for visualization.
"""

from enum import Enum
from pathlib import Path
from typing import Literal, Optional, Dict, Union

import nibabel as nib
import numpy as np

from flask import request
from nibabel.cifti2.cifti2_axes import BrainModelAxis
from werkzeug.datastructures import FileStorage

from findviz.viz import exception
from findviz.viz.io import utils
from findviz.viz.io import validate
from findviz.viz.io.gifti import read_gii, GiftiDict, GiftiFiles
from findviz.viz.transforms import array_to_gifti

# Type aliases
FilePath = Union[str, Path]
FileUploadDict = Dict[str, Optional[Union[FilePath, 'FileStorage']]]

# Expected file upload inputs
class CiftiFiles(Enum):
    DTSERIES = 'cifti_dtseries'
    LEFT_MESH = 'left_gii_mesh'
    RIGHT_MESH = 'right_gii_mesh'

# Browser fields type definition
browser_fields: Dict[str, str] = {
    CiftiFiles.DTSERIES.value: 'cifti-dtseries',
    CiftiFiles.LEFT_MESH.value: 'cifti-surf-left',
    CiftiFiles.RIGHT_MESH.value: 'cifti-surf-right'
}

# Cifti left/right hemisphere axis keys
CIFTI_LEFT_CORTEX = 'CIFTI_STRUCTURE_CORTEX_LEFT'
CIFTI_RIGHT_CORTEX = 'CIFTI_STRUCTURE_CORTEX_RIGHT'

class CiftiUpload:
    """
    Class for handling cifti file uploads.

    Attributes
    ----------
    method : Literal['cli', 'browser']
        The method used for file uploads
    """

    def __init__(
        self,
        method: Literal['cli', 'browser']
    ):
        if method not in ['cli', 'browser']:
            raise ValueError('unrecognized upload method')
        self.method = method

    def upload(
        self,
        fmri_files: Optional[FileUploadDict] = None
    ) -> GiftiDict:
        """
        Upload cifti files

        Parameters
        ----------
        fmri_files : Optional[Dict[str, FilePath]]
            Dictionary of Cifti file paths: 
            {'cifti_dtseries': path, 'left_gii_mesh': path, 'right_gii_mesh': path}
        """
        if self.method == 'cli':
            file_uploads = fmri_files
        elif self.method == 'browser':
            file_uploads = self._get_browser_input()
        
        # check file uploads
        self._check_file_inputs(file_uploads)

        # store whether left and/or right files were uploaded
        if file_uploads[CiftiFiles.LEFT_MESH.value] is None:
            self.left_input = False
        else:
            self.left_input = True

        if file_uploads[CiftiFiles.RIGHT_MESH.value] is None:
            self.right_input = False
        else:
            self.right_input = True
        
        # Initialize output dictionary
        gifti_out = {
            GiftiFiles.LEFT_FUNC.value: None,
            GiftiFiles.RIGHT_FUNC.value: None,
            GiftiFiles.LEFT_MESH.value: None,
            GiftiFiles.RIGHT_MESH.value: None
        }

        # Process dtseries file
        # check dtseries file extension
        if not validate.validate_cii_dtseries_ext(
            utils.get_filename(file_uploads[CiftiFiles.DTSERIES.value])
        ):
            raise exception.FileInputError(
                'File extension is not recognized. Please upload '
                'a cifti file with a .dtseries.nii extension',
                exception.ExceptionFileTypes.CIFTI.value, self.method,
                [browser_fields[CiftiFiles.DTSERIES.value]]
            )
        
        # load and validate dtseries file
        try:
            cifti_img = read_cifti(
                file_uploads[CiftiFiles.DTSERIES.value],
                self.method
            )
            # check if dtseries cifti file has a brainmodel axis
            if not validate.validate_cii_brainmodel_axis(cifti_img):
                raise exception.FileValidationError(
                    'Cifti file does not have a brainmodel axis',
                    validate.validate_cii_brainmodel_axis.__name__,
                    exception.ExceptionFileTypes.CIFTI,
                    [browser_fields[CiftiFiles.DTSERIES.value]]
                )
        except Exception as e:
            raise exception.FileUploadError(
                'Error in reading dtseries cifti file',
                exception.ExceptionFileTypes.CIFTI.value, self.method,
                [browser_fields[CiftiFiles.DTSERIES.value]]
            ) from e

        # Process left hemisphere mesh and extract time series from dtseries file
        if self.left_input:
            if not validate.validate_gii_mesh_ext(
                utils.get_filename(file_uploads[CiftiFiles.LEFT_MESH.value])
            ):
                raise exception.FileInputError(
                    'File extension is not recognized. Please upload '
                    'a gifti file with a .surf.gii extension for left '
                    'hemisphere mesh (geometry) file',
                    exception.ExceptionFileTypes.GIFTI.value, self.method,
                    [browser_fields[CiftiFiles.LEFT_MESH.value]]
                )
            try:
                gii_left_mesh = read_gii(
                    file_uploads[CiftiFiles.LEFT_MESH.value], 
                    self.method
                )
            except Exception as e:
                raise exception.FileUploadError(
                    'Error in reading left hemisphere mesh (geometry) gifti file',
                    exception.ExceptionFileTypes.GIFTI.value, self.method,
                    [browser_fields[CiftiFiles.LEFT_MESH.value]]
                ) from e
                
            if not validate.validate_gii_mesh(gii_left_mesh):
                raise exception.FileValidationError(
                    "The left hemisphere mesh (geometry) file should only contain "
                    "two data arrays, corresponding to coordinate and face arrays."
                    " Check format.",
                    validate.validate_gii_mesh.__name__,
                    exception.ExceptionFileTypes.GIFTI,
                    [browser_fields[CiftiFiles.LEFT_MESH.value]]
                )
            
            gifti_out[GiftiFiles.LEFT_MESH.value] = gii_left_mesh

            # extract left hemisphere time series from dtseries file
            try:
                cifti_left = select_hemisphere_cifti(
                    cifti_img,
                    'left'
                )
                # convert to gifti
                gifti_out[GiftiFiles.LEFT_FUNC.value] = array_to_gifti(
                    cifti_left,
                    both_hemispheres=False
                )
            except exception.FileValidationError as e:
                raise e
            except Exception as e:
                raise exception.FileUploadError(
                    'Error in extracting left hemisphere time series from dtseries file',
                    exception.ExceptionFileTypes.CIFTI.value, self.method,
                    [browser_fields[CiftiFiles.DTSERIES.value]]
                ) from e
            
            # check length of functional and mesh files
            if not validate.validate_gii_func_mesh_len(
                gifti_out[GiftiFiles.LEFT_FUNC.value],
                gifti_out[GiftiFiles.LEFT_MESH.value]
            ):
                raise exception.FileValidationError(
                    "The left hemisphere time course from the dtseries.nii file and mesh (geometry) file "
                    f"are not the same length - {len(gifti_out[GiftiFiles.LEFT_FUNC.value].darrays[0].data)} != "
                    f"{len(gifti_out[GiftiFiles.LEFT_MESH.value].darrays[0].data)}. Check files are consistent "
                    "in length.",
                    validate.validate_gii_func_mesh_len.__name__,
                    exception.ExceptionFileTypes.GIFTI.value,
                    [
                        browser_fields[CiftiFiles.DTSERIES.value], 
                        browser_fields[CiftiFiles.LEFT_MESH.value]
                    ]
                )
        
        # Process right hemisphere mesh file
        if self.right_input:
            if not validate.validate_gii_mesh_ext(
                utils.get_filename(file_uploads[CiftiFiles.RIGHT_MESH.value])
            ):
                raise exception.FileInputError(
                    'File extension is not recognized. Please upload '
                    'a gifti file with a .surf.gii extension for right '
                    'hemisphere mesh (geometry) file',
                    exception.ExceptionFileTypes.GIFTI.value, self.method,
                    [browser_fields[CiftiFiles.RIGHT_MESH.value]]
                )
                
            try:
                gii_right_mesh = read_gii(
                    file_uploads[CiftiFiles.RIGHT_MESH.value],
                    self.method
                )
            except Exception as e:
                raise exception.FileUploadError(
                    'Error in reading right hemisphere mesh (geometry) gifti file',
                    exception.ExceptionFileTypes.GIFTI.value, self.method,
                    [browser_fields[CiftiFiles.RIGHT_MESH.value]]
                ) from e
                
            if not validate.validate_gii_mesh(gii_right_mesh):
                raise exception.FileValidationError(
                    "The right hemisphere mesh (geometry) file should only contain "
                    "two data arrays, corresponding to coordinate and face arrays."
                    " Check format.",   
                    validate.validate_gii_mesh.__name__,
                    exception.ExceptionFileTypes.GIFTI,
                    [browser_fields[CiftiFiles.RIGHT_MESH.value]]
                )
            
            gifti_out[GiftiFiles.RIGHT_MESH.value] = gii_right_mesh
            
            # extract right hemisphere time series from dtseries file
            try:
                cifti_right = select_hemisphere_cifti(
                    cifti_img,
                    'right'
                )
                # convert to gifti
                gifti_out[GiftiFiles.RIGHT_FUNC.value] = array_to_gifti(
                    cifti_right,
                    both_hemispheres=False
                )
            except exception.FileValidationError as e:
                raise e
            except Exception as e:
                raise exception.FileUploadError(
                    'Error in extracting right hemisphere time series from dtseries file',
                    exception.ExceptionFileTypes.CIFTI.value, self.method,
                    [browser_fields[CiftiFiles.DTSERIES.value]]
                ) from e
            
            # check length of functional and mesh files
            if not validate.validate_gii_func_mesh_len(
                gifti_out[GiftiFiles.RIGHT_FUNC.value],
                gifti_out[GiftiFiles.RIGHT_MESH.value]
            ):
                raise exception.FileValidationError(
                    "The right hemisphere time course from the dtseries.nii file and mesh (geometry) file "
                    f"are not the same length - {len(gifti_out[GiftiFiles.RIGHT_FUNC.value].darrays[0].data)} != "
                    f"{len(gifti_out[GiftiFiles.RIGHT_MESH.value].darrays[0].data)}. Check files are consistent "
                    "in length.",
                    validate.validate_gii_func_mesh_len.__name__,
                    exception.ExceptionFileTypes.GIFTI.value,
                    [
                        browser_fields[CiftiFiles.DTSERIES.value], 
                        browser_fields[CiftiFiles.RIGHT_MESH.value]
                    ]
                )
        
        return gifti_out
    
    @staticmethod
    def _get_browser_input() -> FileUploadDict:
        """
        Get file uploads from browser request.

        Returns
        -------
        FileUploadDict
            Dictionary containing file upload objects
        """
        dtseries_file = request.files.get(CiftiFiles.DTSERIES.value)
        left_mesh_file = request.files.get(CiftiFiles.LEFT_MESH.value)
        right_mesh_file = request.files.get(CiftiFiles.RIGHT_MESH.value)

        file_upload = {
            CiftiFiles.DTSERIES.value: dtseries_file,
            CiftiFiles.LEFT_MESH.value: left_mesh_file,
            CiftiFiles.RIGHT_MESH.value: right_mesh_file
        }
        return file_upload
    
    def _check_file_inputs(self, file_inputs: FileUploadDict) -> None:
        """
        Check cifti file inputs.

        Parameters
        ----------
        file_inputs : FileUploadDict
            Dictionary containing file uploads

        Raises
        ------
        FileValidationError
            If file inputs are incorrect or inconsistent
        """
        # get files from dict
        dtseries = file_inputs[CiftiFiles.DTSERIES.value]
        left_mesh = file_inputs[CiftiFiles.LEFT_MESH.value]
        right_mesh = file_inputs[CiftiFiles.RIGHT_MESH.value]

        msg, passed = validate.validate_cii_file_inputs(
            dtseries, 
            left_mesh, 
            right_mesh
        )
        # raise file validation error if file inputs incorrect
        if not passed:
            raise exception.FileValidationError(
                msg, exception.ExceptionFileTypes.CIFTI.value, self.method,
                [
                    browser_fields[CiftiFiles.DTSERIES.value], 
                    browser_fields[CiftiFiles.LEFT_MESH.value],
                    browser_fields[CiftiFiles.RIGHT_MESH.value],
                ]
            )


def read_cifti(
    file: Union[FilePath, FileStorage],
    method: Literal['cli', 'browser']
) -> nib.Cifti2Image:
    """
    Read dense time series cifti (dtseries.nii) file

    Parameters
    ----------
    file : Union[FilePath, FileStorage]
        Path to cifti file
    method : Literal['cli', 'browser']
    """
    if method =='cli':
        try:
            cifti_img = nib.load(file)
        # raise generic exception to be handled higher in stack
        except Exception as e:
            raise e
    elif method == 'browser':
        try:
            file_bytes = file.read()
            cifti_img = nib.Cifti2Image.from_bytes(file_bytes)
        except Exception as e:
            raise e
    
    return cifti_img


def select_hemisphere_cifti(
    cifti_img: nib.Cifti2Image,
    hemisphere: Literal['left', 'right']
) -> np.ndarray:
    """
    Get hemisphere time series from cifti file

    Parameters
    ----------
    cifti_img : nib.Cifti2Image
        Cifti image
    hemisphere : Literal['left', 'right']
        Hemisphere to extract time series from

    Returns
    -------
    np.ndarray
        Time series data

    Raises
    ------
    ValueError
        If hemisphere is not left or right
    FileValidationError
        If cifti image has more than one brainmodel axis
    """
    # get index of brain model axis (previously validated)
    brain_model_idx = [
        i for i in range(cifti_img.ndim) 
        if isinstance(cifti_img.header.get_axis(i), BrainModelAxis)
    ]
    # check if cifti image has more than one brainmodel axis
    if len(brain_model_idx) > 1:
        raise exception.FileValidationError(
            'Cifti dtseries.nii file has more than one brainmodel axis',
            validate.validate_cii_brainmodel_axis.__name__,
            exception.ExceptionFileTypes.CIFTI,
            [browser_fields[CiftiFiles.DTSERIES.value]]
        )
    # get axis key for hemisphere
    if hemisphere == 'left':
        axis_key = CIFTI_LEFT_CORTEX
    elif hemisphere == 'right':
        axis_key = CIFTI_RIGHT_CORTEX
    else:
        raise ValueError('Hemisphere must be either left or right')
    
    # iterate through brain model axis and check if left or right hemisphere axis key matches
    brain_model_axis = cifti_img.header.get_axis(brain_model_idx[0])
    if not validate.validate_cii_hemisphere(
        brain_model_axis,
        hemisphere,
        axis_key
    ):
        raise exception.FileValidationError(
            f'Cifti dtseries.nii file does not have a brainmodel axis for the {hemisphere} '
            f'hemisphere - {axis_key}',
            validate.validate_cii_hemisphere.__name__,
            exception.ExceptionFileTypes.CIFTI,
            [browser_fields[CiftiFiles.DTSERIES.value]]
        )
    # get index of hemisphere time series
    hemisphere_model = [
        bm for bm in brain_model_axis.iter_structures()
        if bm[0] == axis_key
    ][0]
    # slice cifti image to get hemisphere time series
    cifti_data = cifti_img.get_fdata()[:, hemisphere_model[1]]
    # get vertex indices (excludes medial wall)
    vertex_indx = hemisphere_model[2].vertex
    # initialize time series array
    hemisphere_data = np.zeros(
        (vertex_indx.max() + 1, cifti_data.shape[0]),
        dtype=cifti_data.dtype
    ).T
    # iterate through time series and copy data to ts_data
    hemisphere_data[:, vertex_indx] = cifti_data
    # transpose to time points along rows
    return hemisphere_data


"""
This module defines the visualization state classes for the FIND viewer.
"""

from dataclasses import dataclass, field
from typing import Literal, Optional, Dict, Tuple, List, Union

import nibabel as nib
import numpy as np

from findviz.viz.viewer.types import (
    ColorOptions, SliceCoordsDict, OrthoSliceIndexDict, MontageSliceIndexDict,
    MontageSliceCoordsDict, NiftiDataDict, NiftiDataPreprocessedDict,
    GiftiDataDict, GiftiDataPreprocessedDict
)
from findviz.viz.viewer.state.components import (
    FmriPlotOptions, TimeCoursePlotOptions, TaskDesignPlotOptions,
    AnnotationMarkerPlotOptions, TimeMarkerPlotOptions, DistancePlotOptions,
    TimeCourseColor, TimeCourseGlobalPlotOptions
)

@dataclass
class VisualizationState:
    """Base visualization state class containing common attributes.
    
    Attributes:
        tr: Repetition time. If not provided, set to None.
        slicetime_ref: Slicetime reference. If not provided, set to None.
        timepoints: Timepoint indices. Default is None
        timepoints_seconds: Timepoint indices in seconds. Default is None
        global_min: Global minimum value across all timepoints. Default is None
        global_max: Global maximum value across all timepoints. Default is None
        file_type: Type of visualization data ('nifti' or 'gifti'). Default is None
        timepoint: Current timepoint index. Default is 0
        fmri_plot_options: Plot options for fMRI data. Default is FmriPlotOptions()
        preprocessed_fmri_plot_options: Plot options for preprocessed fMRI data. 
            Default is FmriPlotOptions()
        color_options_original: original color options from raw data. 
            Default is ColorOptions()
        preprocessed_color_options_original: original color options from preprocessed data. 
            Default is ColorOptions()
        ts_enabled: Whether time course data is enabled. Default is False
        task_enabled: Whether task design data is enabled. Default is False
        fmri_preprocessed: Whether preprocessed fMRI data is available. Default is False
        ts_preprocessed: Whether preprocessed timecourse data is available. Default is False
        used_colors: Set of used colors for time series. Default is empty set {}
        time_course_global_plot_options: Global time course plot options. 
            Default is TimeCourseGlobalPlotOptions()
        ts_data: Dictionary of timeseries data. Default is empty dict {}
        ts_data_preprocessed: Dictionary of preprocessed timeseries data. 
            Default is None
        ts_labels: Labels for timeseries. Default is empty list [].
        ts_type: Type of timeseries - fmri or user. Default is empty dict {}.
        ts_labels_preprocessed: Labels for preprocessed timeseries. 
            Default is empty list [].
        ts_fmri_plotted: Whether fmri time course is plotted. Default is False.
        ts_plot_options: Dictionary of time course plot options.
            Default is empty dict {}
        task_data: Task design information. Default is None
        task_plot_options: Dictionary of task design plot options. 
            Default is empty dict {}
        annotation_markers: List of annotation markers. Default is empty list []
        annotation_selection: Selected annotation marker. Default is None
        annotation_marker_plot_options: Marker plot options. Default is MarkerPlotOptions()
        time_marker_plot_options: Time marker plot options. Default is TimeMarkerPlotOptions()
        distance_data_enabled: Whether distance data is enabled and plotted. Default is False
        distance_data: Distance data. Default is None
        distance_plot_options: Distance plot options. Default is None.
    """
    # metadata
    tr: Optional[float] = None
    slicetime_ref: Optional[float] = None
    timepoints: Optional[List[int]] = None
    timepoints_seconds: Optional[List[float]] = None
    global_min: Optional[float] = None
    global_max: Optional[float] = None
    file_type: Optional[Literal['nifti', 'gifti']] = None
    timepoint: int = 0

    # plot state
    fmri_plot_options: FmriPlotOptions = field(default_factory=FmriPlotOptions)
    preprocessed_fmri_plot_options: FmriPlotOptions = field(
        default_factory=FmriPlotOptions
    )
    
    # preserved color state
    color_options_original: ColorOptions = field(default_factory=ColorOptions)
    preprocessed_color_options_original: ColorOptions = field(
        default_factory=ColorOptions
    )

    # state flags
    ts_enabled: bool = False
    task_enabled: bool = False
    fmri_preprocessed: bool = False
    ts_preprocessed: Dict[str, bool] = field(default_factory=dict)

    # time course and task design
    used_colors: set[TimeCourseColor] = field(default_factory=set)
    time_course_global_plot_options: TimeCourseGlobalPlotOptions = field(
        default_factory=TimeCourseGlobalPlotOptions
    )
    ts_data: Dict[str, List[float]] = field(default_factory=dict)
    # set ts_labels as a private property
    _ts_labels: List[str] = field(default_factory=list)
    ts_type: Dict[str, Literal['fmri', 'user']] = field(default_factory=dict)
    ts_data_preprocessed: Dict[str, Union[List[float], None]] = field(default_factory=dict)
    ts_labels_preprocessed: List[str] = field(default_factory=list)
    ts_fmri_plotted: bool = False
    ts_plot_options: dict[str, TimeCoursePlotOptions] = field(default_factory=dict)
    task_data: Dict[str, Dict[Literal['block', 'hrf'], List[float]]] = field(default_factory=dict)
    task_conditions: List[str] = field(default_factory=list)
    task_plot_options: dict[str, TaskDesignPlotOptions] = field(default_factory=dict)
    annotation_markers: List[int] = field(default_factory=list)
    annotation_selection: Optional[int] = None
    annotation_marker_plot_options: AnnotationMarkerPlotOptions = field(
        default_factory=AnnotationMarkerPlotOptions
    )
    time_marker_plot_options: TimeMarkerPlotOptions = field(
        default_factory=TimeMarkerPlotOptions
    )
    # outputs from distance analysis
    distance_data_enabled: bool = False
    distance_data: Optional[np.ndarray] = None
    distance_plot_options: Optional[DistancePlotOptions] = None

    @property
    def ts_labels(self) -> List[str]:
        """Get the time series labels."""
        return self._ts_labels

    @ts_labels.setter
    def ts_labels(self, value: List[str]) -> None:
        """Set the time series labels and update preprocessing state."""
        # get the current ts_labels
        current_ts_labels = self._ts_labels
        # set the new ts_labels
        self._ts_labels = value
        # get the newly added ts_labels
        new_ts_labels = set(value) - set(current_ts_labels)
        # get the newly removed ts_labels
        removed_ts_labels = set(current_ts_labels) - set(value)
        # set preprocess state for the new ts_labels
        for ts_label in new_ts_labels:
            self.ts_preprocessed[ts_label] = False
            # set the ts_data_preprocessed to None
            self.ts_data_preprocessed[ts_label] = None

        # remove the removed ts_labels from preprocess state
        for ts_label in removed_ts_labels:
            self.ts_preprocessed.pop(ts_label)
            self.ts_data_preprocessed.pop(ts_label)
    

@dataclass
class NiftiVisualizationState(VisualizationState):
    """Visualization state for NIFTI data.
    
    Attributes:
        file_type: constant 'nifti'
        slice_len: Length of each dimension for NIFTI images. Default is None
        coord_labels: Voxels coordinate labels for NIFTI data. Default is None
        anat_input: Whether anatomical data was provided. Default is False
        mask_input: Whether mask data was provided. Default is False
        nifti_data: Dictionary of NIFTI images. Default is empty dict {}
        nifti_data_preprocessed: Dictionary of preprocessed NIFTI images. 
            Default is empty dict {}
        func_header: Header of functional data. Default is None
        func_affine: Affine matrix of functional data. Default is None
        view_state: View state ('ortho' or 'montage'). Default is 'ortho'
        ortho_slice_idx: Indices of ortho slices for NIFTI data. Default is empty dict {}.
        montage_slice_idx: Indices of montage slices for NIFTI data for each slice direction. 
            Default is empty dict {}.
        ortho_slice_coords: Coordinates of ortho slices for NIFTI data. 
            Default is empty dict {}.
        montage_slice_dir: Direction of montage slice for NIFTI data. 
            Default is 'z'.
        montage_slice_coords: Coordinates of montage slices for NIFTI data. 
            Default is empty dict {}.
        selected_slice: Selected slice for NIFTI data. Default is 'slice_1'.
    """
    # metadata
    file_type: Literal['nifti'] = 'nifti'
    slice_len: Optional[Dict[str, int]] = None # {'x': int, 'y': int, 'z': int}
    coord_labels: Optional[List[Tuple[int, int, int]]] = None # [(x, y, z), (x, y, z), (x, y, z)]
    anat_input: bool = False
    mask_input: bool = False

    # nifti data
    nifti_data: NiftiDataDict = field(default_factory=dict)
    nifti_data_preprocessed: NiftiDataPreprocessedDict = field(default_factory=dict)

    # functional header and affine
    func_header: Optional[nib.Nifti1Header] = None
    func_affine: Optional[np.ndarray] = None

    # view state - orthogonal or montage view
    view_state: Literal['ortho', 'montage'] = 'ortho'

    # orthogonal view slice indices
    ortho_slice_idx: OrthoSliceIndexDict = field(default_factory=dict)
    # montage view slice indices
    montage_slice_idx: MontageSliceIndexDict = field(default_factory=dict)

    # ortho view slice coordinates
    ortho_slice_coords: Optional[SliceCoordsDict] = field(default_factory=dict)

    # montage view slice coordinates
    montage_slice_dir: Optional[Literal['x', 'y', 'z']] = 'z'
    montage_slice_coords: Optional[MontageSliceCoordsDict] = field(default_factory=dict)

    # selected slice
    selected_slice: Literal['slice_1', 'slice_2', 'slice_3'] = 'slice_1'


@dataclass
class GiftiVisualizationState(VisualizationState):
    """Visualization state for GIFTI data.
    
    Attributes:
        file_type: constant 'gifti'
        left_input: Whether left hemisphere data was provided
        right_input: Whether right hemisphere data was provided
        gifti_data: Dictionary of GIFTI images
        gifti_data_preprocessed: Dictionary of preprocessed GIFTI images
        vertices_left: Vertex coordinates for left hemisphere
        faces_left: Face indices for left hemisphere
        vertices_right: Vertex coordinates for right hemisphere
        faces_right: Face indices for right hemisphere
        coord_labels: Vertex number and hemisphere for each vertex
        selected_vertex: Selected vertex index
        selected_hemi: Selected hemisphere ('left' or 'right')
    """
    # metadata
    file_type: Literal['gifti'] = 'gifti'
    left_input: bool = False
    right_input: bool = False
    vertices_left: Optional[List[float]] = None
    faces_left: Optional[List[float]] = None
    vertices_right: Optional[List[float]] = None
    faces_right: Optional[List[float]] = None
    both_hemispheres: bool = False
    
    # gifti data
    gifti_data: GiftiDataDict = field(default_factory=dict)
    gifti_data_preprocessed: GiftiDataPreprocessedDict = field(default_factory=dict)

    # selected vertex and hemisphere
    selected_vertex: Optional[int] = None
    selected_hemi: Optional[Literal['left', 'right']] = None

    # vertex number and hemisphere for each vertex
    left_coord_labels: Optional[List[Tuple[int, Literal['left']]]] = None
    right_coord_labels: Optional[List[Tuple[int, Literal['right']]]] = None

    def __post_init__(self):
        # set both hemispheres to True if left and right hemispheres are provided
        if self.left_input and self.right_input:
            self.both_hemispheres = True
            


"""Preprocess utilities for functional MRI and time course data"""

from findviz.viz.preprocess import fmri
from findviz.viz.preprocess import input
from findviz.viz.preprocess import timecourse
from findviz.viz.preprocess import utils
from findviz.viz.preprocess import validate


__all__ = ['fmri', 'input', 'timecourse', 'utils', 'validate']



"""Utility functions for processing neuroimaging data.

This module provides helper functions for processing NIFTI and GIFTI data,
including metadata extraction and value range calculations.

Functions:
    - apply_mask_nifti: Apply a mask to a NIfTI image
    - extend_color_range: Extend color range by a given percentage
    - get_coord_labels_gifti: Get coordinate labels for GIFTI data as a list of tuples
    - get_coord_labels_nifti: Get coordinate labels for NIFTI data as a 3D array
    - get_fmri_minmax: Calculate global minimum and maximum values for fmri data
    - get_ts_minmax: Calculate global minimum and maximum values for time series data
    - get_ortho_slice_coords: Get initial orthogonal view slice coordinates for NIFTI data
    - get_ortho_slice_idx: Get initial orthogonal view slice indices for NIFTI data
    - get_montage_slice_coords: Get initial montage slice coordinates for NIFTI data
    - get_montage_slice_idx: Get initial montage slice indices for NIFTI data
    - get_slider_step_size: Calculate slider step size
    - get_precision: Calculate precision for slider step size
    - package_distance_metadata: Package metadata for distance visualization
    - package_gii_metadata: Package metadata for GIFTI visualization
    - package_nii_metadata: Package metadata for NIFTI visualization
    - transform_to_world_coords: Transform voxel coordinates to world coordinates
    - requires_state: Decorator to check if state exists before executing method
"""
import decimal

from functools import wraps
from typing import Dict, Tuple, Union, Optional, List, Literal

import numpy as np
import nibabel as nib

from nibabel.gifti import GiftiImage
from nibabel.nifti1 import Nifti1Image

from findviz.logger_config import setup_logger

logger = setup_logger(__name__)

# define slice containers for nifti visualization
slices_containers = ['slice_1', 'slice_2', 'slice_3']

def apply_mask_nifti(
    nifti_img: nib.Nifti1Image,
    mask_img: nib.Nifti1Image,
) -> nib.Nifti1Image:
    """Apply a mask to a NIfTI image

    Parameters
    ----------
    nifti_img : nib.Nifti1Image
        NIfTI image to mask
    mask_img : nib.Nifti1Image
        Mask image to apply

    Returns
    -------
    nib.Nifti1Image
        Masked NIfTI image
    """
    # apply mask
    nifti_data = nifti_img.get_fdata()
    mask_data = mask_img.get_fdata()
    nifti_data[mask_data == 0, :] = np.nan
    masked_img = nib.Nifti1Image(nifti_data, nifti_img.affine, nifti_img.header)
    return masked_img

def extend_color_range(
    color_min: float,
    color_max: float,
    extend_range: float = 0.25
) -> Tuple[float, float]:
    """Extend color range by a given percentage
    
    Parameters:
    -----------
    color_min: Minimum color value
    color_max: Maximum color value
    extend_range: Percentage to extend color range by
    
    Returns:
    --------
    Tuple containing (extended_color_min, extended_color_max)
    """
    return color_min - (extend_range * abs(color_min)), color_max + (extend_range * abs(color_max))


def get_coord_labels_gifti(
    gifti_img: Optional[GiftiImage] = None,
    hemisphere: Optional[Literal['left', 'right']] = None
) -> List[Tuple[int, Literal['left', 'right']]]:
    """Get coordinate labels for GIFTI data as a list of tuples
    containing the vertex number and hemisphere for each vertex. If
    gifti_img is none, return None.

    Parameters
    ----------
    gifti_img : GiftiImage
        GIFTI image
    hemisphere : Literal['left', 'right']
        Hemisphere to get coordinate labels for

    Returns
    -------
    List[Tuple[int, Literal['left', 'right']]]
        List of tuples containing the vertex number and hemisphere for each vertex
    """
    if gifti_img is None:
        return None
    # get the vertex number and hemisphere for each vertex
    vertex_numbers = range(len(gifti_img.darrays[0].data))
    hemispheres = [hemisphere] * len(vertex_numbers)
    # create a list of tuples containing the vertex number and hemisphere for each vertex
    coord_labels = list(zip(vertex_numbers, hemispheres))
    return coord_labels


def get_coord_labels_nifti(
    nii_img: Nifti1Image
) -> np.ndarray:
    """Get coordinate labels for NIFTI data as a 3D array
    
    Parameters
    ----------
    nii_img : Nifti1Image
        Input NIFTI image
        
    Returns
    -------
    np.ndarray
        3D array of shape (X, Y, Z, 3) containing the coordinate indices
        for each voxel position
    """
    data = nii_img.get_fdata()
    shape = data.shape[:3]  # Get first 3 dimensions (X, Y, Z)
    
    # Create coordinate arrays for each dimension
    x_coords = np.arange(shape[0])[:, np.newaxis, np.newaxis]
    y_coords = np.arange(shape[1])[np.newaxis, :, np.newaxis]
    z_coords = np.arange(shape[2])[np.newaxis, np.newaxis, :]
    
    # Broadcast to full 3D arrays
    X = np.broadcast_to(x_coords, shape)
    Y = np.broadcast_to(y_coords, shape)
    Z = np.broadcast_to(z_coords, shape)
    
    # Stack the coordinates into a single array with shape (X, Y, Z, 3)
    coord_labels = np.stack([X, Y, Z], axis=-1)

    # Vectorize the string formatting function
    format_coord = np.vectorize(lambda x, y, z: f"Voxel: {x}, {y}, {z}")

    # Apply the formatting function to each coordinate
    coord_labels = format_coord(
        coord_labels[..., 0], 
        coord_labels[..., 1], 
        coord_labels[..., 2]
    )
    
    return coord_labels


def get_fmri_minmax(
    data: Union[np.ndarray, GiftiImage], 
    file_type: Literal['gifti', 'nifti']
) -> Tuple[float, float]:
    """Calculate global minimum and maximum values for fmri data.
    
    Parameters:
    -----------
    data : Union[np.ndarray, GiftiImage]
        Input data array or GiftiImage
    file_type : Literal['gifti', 'nifti']
        Type of neuroimaging data ('gifti' or 'nifti')
    
    Returns:
    --------
    Tuple containing (global_minimum, global_maximum)
    
    Raises:
    --------
    ValueError: If file_type is not 'gifti' or 'nifti'
    """
    if file_type == 'nifti':
        if not isinstance(data, np.ndarray):
            raise TypeError("NIFTI data must be numpy array")
        data_min = float(np.nanmin(data))
        data_max = float(np.nanmax(data))
    elif file_type == 'gifti':
        if not isinstance(data, GiftiImage):
            raise TypeError("GIFTI data must be GiftiImage")
        data_min = float(np.nanmin([np.min(d.data) for d in data.darrays]))
        data_max = float(np.nanmax([np.max(d.data) for d in data.darrays]))
    else:
        raise ValueError("file_type must be 'gifti' or 'nifti'")

    return data_min, data_max


def get_ortho_slice_coords(
    ortho_slice_idx: Dict[str, int]
) -> Dict[str, Dict[str, int]]:
    """Get initial orthogonal view slice coordinates for 
    NIFTI data as middle of x,y,z dimensions

    Parameters:
    -----------
    ortho_slice_idx: Dictionary containing x,y,z slice indices
    
    Returns:
    --------
    Dictionary containing x,y,z slice coordinates
    """
    ortho_slice_coords = {}

    for slice in slices_containers:
        # saggital slice
        if slice == 'slice_1':
            ortho_slice_coords[slice] = {
                'x': ortho_slice_idx['y'],
                'y': ortho_slice_idx['z'],
            }
        # coronal slice
        elif slice == 'slice_2':
            ortho_slice_coords[slice] = {
                'x': ortho_slice_idx['x'],
                'y': ortho_slice_idx['z']
            }
        # axial slice
        elif slice == 'slice_3':
            ortho_slice_coords[slice] = {
                'x': ortho_slice_idx['x'],
                'y': ortho_slice_idx['y']
            }
    return ortho_slice_coords


def get_ortho_slice_idx(slice_len: Dict[str, int]) -> Dict[str, int]:
    """Get orthogonal view slice indices for NIFTI data. Set as middle of x,y,z dimensions
    
    Parameters:
    -----------
    slice_len: Dictionary containing x,y,z dimensions
    
    Returns:
    --------
    Dictionary containing x,y,z slice indices
    """
    return {
        'x': int(slice_len['x'] / 2),
        'y': int(slice_len['y'] / 2),
        'z': int(slice_len['z'] / 2)
    }


def get_montage_slice_coords(
    ortho_slice_coords: Dict[str, Dict[str, int]],
    montage_slice_dir: Literal['x', 'y', 'z']
) -> Dict[str, Dict[str, int]]:
    """Get montage slice coordinates for NIFTI data for a given montage slice direction 
    from orthogonal slice coordinates. Initialize montage click coordinates with 
    orthogonal slice coordinates for the corresponding slice direction.

    Parameters:
    -----------
    ortho_slice_coords: Dictionary containing x, y slice coordinates for orthogonal view
    montage_slice_dir: Direction of montage slices

    Returns:
    --------
    Dictionary containing x, y click coordinates for montage view
    """
    if montage_slice_dir == 'x':
        x = ortho_slice_coords['slice_1']['x']
        y = ortho_slice_coords['slice_2']['y']
    elif montage_slice_dir == 'y':
        x = ortho_slice_coords['slice_2']['x']
        y = ortho_slice_coords['slice_3']['y']
    elif montage_slice_dir == 'z':
        x = ortho_slice_coords['slice_3']['x']
        y = ortho_slice_coords['slice_1']['y']

    montage_slice_coords = {}
    for slice in slices_containers:
        montage_slice_coords[slice] = {
            'x': x,
            'y': y,
        }
    return montage_slice_coords


def get_montage_slice_idx(
    slice_len: Dict[str, int],
    montage_slice_dir: Literal['x', 'y', 'z'],
    ortho_slice_idx: Dict[str, int]
) -> Dict[str, Dict[str, int]]:
    """Get initial montage slice indices for NIFTI data as 
    equally distributed across the slice direction (montage_slice_dir). 
    The montage slice indices are initialized with the slice indices
    for each slice direction, and placed at 33%, 50%, and 66% of the slice length
    in the montage slice direction.

    Parameters:
    -----------
    slice_len: Dictionary containing x,y,z dimensions
    montage_slice_dir: Direction of montage slices
    ortho_slice_idx: Dictionary containing x,y,z slice indices

    Returns:
    --------
    Dictionary containing montage slice indices
    """
    # percentage of slice length to start montage slice at
    slice_len_dir = slice_len[montage_slice_dir]
    montage_slice_perc = [0.33, 0.5, 0.66]
    montage_slice_idx = {}
    for perc, slice in zip(montage_slice_perc, slices_containers):
        slice_idx_dir = int(slice_len_dir * perc)
        # initialize montage slice indices with orthogonal slice indices
        montage_slice_idx[slice] = {
            'x': ortho_slice_idx['x'],
            'y': ortho_slice_idx['y'],
            'z': ortho_slice_idx['z']
        }
        # set montage slice index for current slice direction
        montage_slice_idx[slice][montage_slice_dir] = slice_idx_dir

    return montage_slice_idx


def get_precision(
    data_range: float,
    max_precision: int = 6
) -> int:
    """Calculate precision for slider step size"""
    # Handle NaN values
    if np.isnan(data_range):
        return 0
        
    data_range_dec = decimal.Decimal(str(data_range))
    precision = abs(data_range_dec.as_tuple().exponent)
    if precision > max_precision:
        return max_precision
    return precision


def get_slider_step_size(
    data_range: float,
    slider_steps: int,
    precision: int
) -> float:
    """Calculate slider step size
    
    Parameters:
    -----------
    data_range: Range of data
    slider_steps: Number of steps in slider
    precision: Precision of step size
    
    Returns:
    --------
    Slider step size
    """
    return round(data_range/slider_steps, precision)


def get_ts_minmax(
    default_min: float,
    default_max: float,
    ts_data: Optional[Dict[str, np.ndarray]] = None,
    task_data: Optional[Dict[str, np.ndarray]] = None
) -> Tuple[float, float]:
    """Calculate global minimum and maximum values for time series data. If 
    no ts data or task data is provided, return default_min and default_max.
    
    Parameters:
    -----------
    default_min: Default minimum value
    default_max: Default maximum value
    ts_data: Dictionary containing time series data
    task_data: Dictionary containing task data
    
    Returns:
    --------
    Tuple containing (global_minimum, global_maximum)
    """
    ts_min = np.nan
    ts_max = np.nan
    task_min = np.nan
    task_max = np.nan

    if ts_data is not None:
        ts_min = float(np.nanmin([ts_data[ts] for ts in ts_data]))
        ts_max = float(np.nanmax([ts_data[ts] for ts in ts_data]))

    if task_data is not None:
        task_min = float(np.nanmin([task_data[task] for task in task_data]))
        task_max = float(np.nanmax([task_data[task] for task in task_data]))

    global_min = float(np.nanmin([ts_min, task_min]))
    global_max = float(np.nanmax([ts_max, task_max]))

    if np.isnan(global_min):
        global_min = default_min
    if np.isnan(global_max):
        global_max = default_max

    return global_min, global_max


def package_distance_metadata(
    distance_data: np.ndarray,
    slider_steps: int = 100,
    precision: int = 6
) -> Dict[str, Union[float, List[int]]]:
    """Package metadata for distance visualization.
    
    Parameters:
    -----------
    distance_data: Distance data
    slider_steps: Number of steps in slider
    precision: Precision of step size
    
    Returns:
    --------
    Dictionary containing:
        - color_min: Minimum value for color mapping
        - color_max: Maximum value for color mapping
        - color_range: Color range for color mapping
        - precision: Precision of the color mapping
        - slider_step_size: Stepsize of the sliders
    """
    # Calculate precision for slider step size
    precision = get_precision(
        data_range=distance_data.max() - distance_data.min()
    )
    metadata: Dict[str, Union[float, List[int]]] = {
        'color_min': distance_data.min(),
        'color_max': distance_data.max(),
        'color_range': extend_color_range(
            distance_data.min(),
            distance_data.max()
        ),
        'precision': precision,
    }
    metadata['slider_step_size'] = get_slider_step_size(
        metadata['color_range'][1] - metadata['color_range'][0],
        slider_steps,
        metadata['precision']
    )
    return metadata


def package_gii_metadata(
    left_img: Optional[GiftiImage],
    right_img: Optional[GiftiImage],
    slider_steps: int = 100,
    precision: int = 6
) -> Dict[str, Union[float, List[int]]]:
    """Package metadata for GIFTI visualization.
    
    Parameters:
    -----------
    left_img: Left hemisphere GIFTI image
    right_img: Right hemisphere GIFTI image
    slider_steps: Number of steps in slider
    precision: Precision of step size
    
    Returns:
    --------
    Dictionary containing:
        - global_min: Global minimum value across both hemispheres
        - global_max: Global maximum value across both hemispheres
        - timepoints: List of timepoint indices
        - color_min: Minimum value for color mapping
        - color_max: Maximum value for color mapping
        - color_range: Color range for color mapping
        - precision: Precision of the color mapping
        - slider_step_size: Stepsize of the sliders
    """
    metadata: Dict[str, Union[float, List[int]]] = {}
    timepoints: Optional[List[int]] = None
    
    # Initialize with NaN
    global_min_left = np.nan
    global_max_left = np.nan
    global_min_right = np.nan
    global_max_right = np.nan
    
    if left_img is not None:
        global_min_left, global_max_left = get_fmri_minmax(left_img, 'gifti')
        timepoints = list(range(len(left_img.darrays)))
        
    if right_img is not None:
        global_min_right, global_max_right = get_fmri_minmax(right_img, 'gifti')
        if timepoints is None:
            timepoints = list(range(len(right_img.darrays)))

    metadata['global_min'] = float(np.nanmin([global_min_left, global_min_right]))
    metadata['global_max'] = float(np.nanmax([global_max_left, global_max_right]))
    # Calculate precision for slider step size
    precision = get_precision(
        data_range=metadata['global_max'] - metadata['global_min']
    )
    metadata['precision'] = precision
    metadata['timepoints'] = timepoints or []
    metadata['color_min'] = metadata['global_min']
    metadata['color_max'] = metadata['global_max']
    metadata['color_range'] = extend_color_range(
        metadata['color_min'], 
        metadata['color_max']
    )
    metadata['threshold_range'] = metadata['color_range']
    metadata['slider_step_size'] = get_slider_step_size(
        metadata['color_range'][1] - metadata['color_range'][0],
        slider_steps,
        metadata['precision']
    )
    
    return metadata


def package_nii_metadata(
    nii_img: Nifti1Image,
    slider_steps: int = 100,
    precision: int = 6
) -> Dict[str, Union[float, List[int], Dict[str, int]]]:
    """Package metadata for NIFTI visualization.
    
    Parameters:
    -----------
    nii_img: NIFTI image object
    
    Returns:
    --------
    Dictionary containing:
        - global_min: Global minimum value
        - global_max: Global maximum value
        - color_min: Minimum value for color mapping
        - color_max: Maximum value for color mapping
        - color_range: Color range for color mapping
        - timepoints: List of timepoint indices
        - slice_len: Dictionary of x,y,z dimensions
        - ortho_slice_idx: Dictionary of x,y,z slice indices for orthogonal view
        - ortho_slice_coords: Dictionary of x,y,z slice coordinates for orthogonal view
        - montage_slice_idx: Dictionary of x,y,z slice indices for montage view
        - montage_slice_coords: Dictionary of x,y,z slice coordinates for montage view
        - threshold_range: Threshold range for threshold mapping
        - precision: Precision of the color mapping
        - slider_step_size: Stepsize of the sliders
    """
    data = nii_img.get_fdata()
    # Calculate global min and max
    data_min, data_max = get_fmri_minmax(data, 'nifti')
    # Calculate precision for slider step size
    precision = get_precision(data_range=data_max - data_min)
    # Get initial orthogonal view slice indices
    slice_len = {
        'x': int(data.shape[0]),
        'y': int(data.shape[1]),
        'z': int(data.shape[2])
    }
    # Get initial orthogonal view slice indices
    ortho_slice_idx = get_ortho_slice_idx(slice_len)
    # Get initial orthogonal view slice coordinates
    ortho_slice_coords = get_ortho_slice_coords(ortho_slice_idx)
    # Get initial montage slice indices
    montage_slice_idx_x = get_montage_slice_idx(slice_len, 'x', ortho_slice_idx)
    montage_slice_idx_y = get_montage_slice_idx(slice_len, 'y', ortho_slice_idx)
    montage_slice_idx_z = get_montage_slice_idx(slice_len, 'z', ortho_slice_idx)
    montage_slice_idx = {
        'x': montage_slice_idx_x,
        'y': montage_slice_idx_y,
        'z': montage_slice_idx_z
    }
    # Get initial montage slice coordinates
    montage_slice_coords_x = get_montage_slice_coords(ortho_slice_coords, 'x')
    montage_slice_coords_y = get_montage_slice_coords(ortho_slice_coords, 'y')
    montage_slice_coords_z = get_montage_slice_coords(ortho_slice_coords, 'z')
    montage_slice_coords = {
        'x': montage_slice_coords_x,
        'y': montage_slice_coords_y,
        'z': montage_slice_coords_z
    }

    metadata: Dict[str, Union[float, List[int], Dict[str, int]]] = {
        'global_min': float(data_min),
        'global_max': float(data_max),
        'color_min': data_min,
        'color_max': data_max,
        'color_range': extend_color_range(data_min, data_max),
        'timepoints': list(range(data.shape[3])) if len(data.shape) > 3 else [0],
        'slice_len': slice_len,
        'ortho_slice_idx': ortho_slice_idx,
        'ortho_slice_coords': ortho_slice_coords,
        'montage_slice_idx': montage_slice_idx,
        'montage_slice_coords': montage_slice_coords,
        'threshold_range': extend_color_range(data_min, data_max),
        'precision': precision,
        'slider_step_size': get_slider_step_size(
            data_max - data_min,
            slider_steps,
            precision
        )
    }
    
    return metadata


def transform_to_world_coords(
    voxel_coords: Dict[Literal['x', 'y', 'z'], int],
    affine: np.ndarray
) -> Tuple[float, float, float]:
    """Transform voxel coordinates to world coordinates.
    
    Parameters:
    -----------
    voxel_coords: Dictionary of x,y,z voxel coordinates
    affine: Affine matrix
    
    Returns:
    --------
    Tuple of x,y,z world coordinates
    """
    coords_array = np.array(
        [voxel_coords['x'], voxel_coords['y'], voxel_coords['z'], 1]
    )
    return np.dot(affine, coords_array)[:3]


def requires_state(func):
    """Decorator to check if state exists before executing method.
    
    Checks if self._state has been initialized before executing the decorated method.
    If no state exists, logs an error and returns None.
    """
    # For regular methods
    if not isinstance(func, property):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self._state:
                logger.error(f"No state exists. Must call create_nifti_state or create_gifti_state before {func.__name__}")
                return None
            return func(self, *args, **kwargs)
        return wrapper
    
    # For properties
    else:
        @property
        @wraps(func)
        def wrapper(self):
            if not self._state:
                logger.error(f"No state exists. Must call create_nifti_state or create_gifti_state before {func.fget.__name__}")
                return None
            return func.fget(self)
        return wrapper






"""
Routes for display of color maps
"""
from enum import Enum

import plotly.colors as pc

from flask import Blueprint
 
from findviz.logger_config import setup_logger
from findviz.routes.utils import (
    handle_route_errors, 
    Routes
)
from findviz.viz.viewer.state.components import ColorMaps
from findviz.routes.shared import data_manager

# Set up a logger for the app
logger = setup_logger(__name__)

color_bp = Blueprint('color', __name__)

# Route to provide colormap data
@color_bp.route(Routes.GET_COLORMAPS.value, methods=['GET'])
@handle_route_errors(
    error_msg='Error generating colormap data',
    log_msg='Generated colormap data successfully',
    fmri_file_type=lambda: data_manager.ctx.fmri_file_type,
    route=Routes.GET_COLORMAPS
)
def get_colormaps() -> dict:
    """Get colormap data"""
    return generate_colormap_data(ColorMaps)


def generate_colormap_data(
    colormaps: list[ColorMaps]
) -> dict[str, dict[str, str]]:
    """
    Generate colormap data

    Arguments
    ---------
    colormaps : list[ColorMaps]
        List of color maps

    Returns
    -------
    dict[str, dict[str, str]]
        Colormap data
    """
    colormap_data = {}
    for cmap in colormaps:
        # Get the colorscale for each colormap
        colormap_colors = pc.get_colorscale(cmap.value)
        # Convert Plotly color codes to rgb
        rgb_colors = [
            (pos, code_to_rgb(color_code))
            for pos, color_code in colormap_colors
        ]
        # Put into html
        gradient = "linear-gradient(to right, " + ", ".join(
            [f"rgb({r},{g},{b})" for pos, (r, g, b) in rgb_colors]
        ) + ")"
        # package output
        colormap_data[cmap.value] = {
            'label': cmap.value,
            'gradient': gradient
        }
    return colormap_data


def code_to_rgb(color_code: str) -> list[int]:
    """
    Convert color code to rgb

    Arguments
    ---------
    color_code : str
        Color code

    Returns
    -------
    list[int]
        RGB values
    """
    # if hex format
    if color_code.startswith('#'):
        color_rgb = pc.hex_to_rgb(color_code)
    # otherwise, assume rgb
    else:
        # clean string
        color_clean = color_code.replace('rgb(','').replace(')','')
        color_rgb = [int(c) for c in color_clean.split(',')]
    return color_rgb



"""Utility Modules for route input handling"""
from enum import Enum
from functools import wraps
from typing import Union, Callable, TypeVar, ParamSpec, List, Type


import numpy as np

from flask import make_response, request

from findviz.routes.shared import data_manager
from findviz.viz.exception import DataRequestError
from findviz.logger_config import setup_logger


# Set up logger
logger = setup_logger(__name__)

# Type variables for generic function signatures
P = ParamSpec('P')
R = TypeVar('R')

class Routes(Enum):
    ADD_ANNOTATION_MARKER='/add_annotation_marker'
    CHANGE_TASK_CONVOLUTION='/change_task_convolution'
    CHECK_CACHE='/check_cache'
    CHECK_TS_PREPROCESSED='/check_ts_preprocessed'
    CHECK_FMRI_PREPROCESSED='/check_fmri_preprocessed'
    CLEAR_ANNOTATION_MARKERS='/clear_annotation_markers'
    CLEAR_CACHE='/clear_cache'
    CONVERT_TIMEPOINTS='/convert_timepoints'
    CORRELATE='/correlate'
    DISTANCE='/distance'
    GET_ANNOTATION_MARKERS='/get_annotation_markers'
    GET_ANNOTATION_MARKER_PLOT_OPTIONS='/get_annotation_marker_plot_options'
    GET_CLICK_COORDS='/get_click_coords'
    GET_COORD_LABELS='/get_coord_labels'
    GET_CROSSHAIR_COORDS='/get_crosshair_coords'
    GET_COLORMAPS='/get_colormaps'
    GET_DIRECTION_LABEL_COORDS='/get_direction_label_coords'
    GET_DISTANCE_DATA='/get_distance_data'
    GET_DISTANCE_PLOT_OPTIONS='/get_distance_plot_options'
    GET_FMRI_DATA='/get_fmri_data'
    GET_FMRI_PLOT_OPTIONS='/get_fmri_plot_options'
    GET_HEADER='/get_header'
    GET_LAST_TIMECOURSE='/get_last_timecourse'
    GET_LOG_ENTRIES='/get_log_entries'
    GET_LOG_FILES='/get_log_files'
    GET_MONTAGE_DATA='/get_montage_data'
    GET_NIFTI_VIEW_STATE='/get_nifti_view_state'
    GET_SELECTED_TIME_POINT='/get_selected_time_point'
    GET_TASK_CONDITIONS='/get_task_conditions'
    GET_TASK_DESIGN_PLOT_OPTIONS='/get_task_design_plot_options'
    GET_TIMECOURSE_LABELS='/get_timecourse_labels'
    GET_TIMECOURSE_LABELS_PREPROCESSED='/get_timecourse_labels_preprocessed'
    GET_TIMECOURSE_DATA='/get_timecourse_data'
    GET_TIMECOURSE_GLOBAL_PLOT_OPTIONS='/get_timecourse_global_plot_options'
    GET_TIMECOURSE_SOURCE='/get_timecourse_source'
    GET_TIMECOURSE_PLOT_OPTIONS='/get_timecourse_plot_options'
    GET_TIMECOURSE_SHIFT_HISTORY='/get_timecourse_shift_history'
    GET_TIMEMARKER_PLOT_OPTIONS='/get_timemarker_plot_options'
    GET_TIMEPOINT='/get_timepoint'
    GET_TIMEPOINTS='/get_timepoints'
    GET_TS_FMRI_PLOTTED='/get_ts_fmri_plotted'
    GET_PREPROCESSED_FMRI='/get_preprocessed_fmri'
    GET_PREPROCESSED_TIMECOURSE='/get_preprocessed_timecourse'
    GET_VERTEX_COORDS='/get_vertex_coords'
    GET_VIEWER_METADATA='/get_viewer_metadata'
    GET_VOXEL_COORDS='/get_voxel_coords'
    GET_WORLD_COORDS='/get_world_coords'
    FIND_PEAKS='/find_peaks'
    MOVE_ANNOTATION_SELECTION='/move_annotation_selection'
    POP_FMRI_TIMECOURSE='/pop_fmri_timecourse'
    REMOVE_DISTANCE_PLOT='/remove_distance_plot'
    REMOVE_FMRI_TIMECOURSES='/remove_fmri_timecourses'
    RESET_FMRI_COLOR_OPTIONS='/reset_fmri_color_options'
    RESET_FMRI_PREPROCESS='/reset_fmri_preprocess'
    RESET_TIMECOURSE_PREPROCESS='/reset_timecourse_preprocess'
    RESET_TIMECOURSE_SHIFT='/reset_timecourse_shift'
    SAVE_SCENE='/save_scene'
    TEST_ERROR='/test_error'
    UNDO_ANNOTATION_MARKER='/undo_annotation_marker'
    UPDATE_ANNOTATION_MARKER_PLOT_OPTIONS='/update_annotation_marker_plot_options'
    UPDATE_DISTANCE_PLOT_OPTIONS='/update_distance_plot_options'
    UPDATE_FMRI_TIMECOURSE='/update_fmri_timecourse'
    UPDATE_FMRI_PLOT_OPTIONS='/update_fmri_plot_options'
    UPDATE_MONTAGE_SLICE_DIR='/update_montage_slice_dir'
    UPDATE_MONTAGE_SLICE_IDX='/update_montage_slice_idx'
    UPDATE_NIFTI_VIEW_STATE='/update_nifti_view_state'
    UPDATE_TASK_DESIGN_PLOT_OPTIONS='/update_task_design_plot_options'
    UPDATE_TIMEPOINT='/update_timepoint'
    UPDATE_TIMECOURSE_PLOT_OPTIONS='/update_timecourse_plot_options'
    UPDATE_TIMECOURSE_GLOBAL_PLOT_OPTIONS='/update_timecourse_global_plot_options'
    UPDATE_TIMEMARKER_PLOT_OPTIONS='/update_timemarker_plot_options'
    UPDATE_TIMECOURSE_SHIFT='/update_timecourse_shift'
    UPDATE_LOCATION='/update_location'
    UPDATE_TR='/update_tr'
    UPLOAD_FILES='/upload'
    UPLOAD_SCENE='/upload_scene'
    WINDOWED_AVERAGE='/windowed_average'



# Convert values from json
def convert_value(
    value: Union[str, bool, int, float]
) -> Union[str, int, float, None]:
    """
    Convert values from form data to Python types 

    Args:
        value (Union[str, bool, int, float]): value from web browser

    Returns:
        Union[str, int, float, None]: converted value
    """
    # check if value is a boolean
    if isinstance(value, bool):
        return value
    # Check for booleans passed as strings
    if isinstance(value, str):
        if value.lower() == 'true':
            return True
        elif value.lower() == 'false':
            return False
    # Check for None
    if value.lower() == 'null' or value.lower() == 'none' or value == '':
        return None
    # Try to convert to integer
    try:
        return int(value)
    except ValueError:
        pass
    # Try to convert to float
    try:
        return float(value)
    except ValueError:
        pass
    # Return the value as-is if no conversion happened
    return value


def handle_context() -> Callable[[Callable[P, R]], Callable[P, R]]:
    """Decorator to handle context switching for routes.
    """
    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            # Get context_id from query parameters or form data
            context_id = request.args.get(
                'context_id',
                request.form.get('context_id')
            )
            
            # Switch to the requested context
            try:
                data_manager.switch_context(context_id)
            except ValueError as e:
                logger.error(f"Invalid context requested: {context_id}")
                return make_response({'error': str(e)}, 400)
                
            return func(*args, **kwargs)
        return wrapper
    return decorator

def handle_route_errors(
    error_msg: str,
    log_msg: str = None,
    fmri_file_type: Union[str, Callable[[], str]] = None,  # callable
    route: Enum = None,
    route_parameters: List[str] = None,
    custom_exceptions: List[Type[Exception]] = None,
) -> Callable[[Callable[P, R]], Callable[P, tuple[Union[R, str], int]]]:
    """
    Decorator to handle common route error patterns
    
    Parameters
    ----------
    error_msg : str
        Base error message for the route
    log_msg : str, optional
        Message to log on success
    fmri_file_type : Union[str, Callable[[], str]], optional
        FMRI file type for DataRequestError. Pass a callable to get the fmri file type dynamically
    route : Enum, optional
        Route name as an Enum for DataRequestError
    route_parameters : List[str], optional
        List of required route parameters to check
    custom_exceptions : List[Type[Exception]], optional
        List of custom exceptions to handle

    Returns
    -------
    Callable
        Decorated route function that handles errors consistently

    """
    def decorator(func: Callable[P, R]) -> Callable[P, tuple[Union[R, str], int]]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> tuple[Union[R, str], int]:
             # Get file type at runtime if it's a callable
            current_file_type = fmri_file_type() if callable(fmri_file_type) else fmri_file_type

            # check if route parameters are provided
            if route_parameters:
                # check if all expected route parameters are provided
                for param in route_parameters:
                    if (param not in request.form) and (param not in request.args):
                        # Handle missing required fields
                        data_error = DataRequestError(
                            message=f'{error_msg}.',
                            fmri_file_type=current_file_type,
                            route=route.value,
                            input_field=param
                        )
                        logger.error(data_error)
                        return make_response(data_error.message, 400)

            try:
                # Execute the route function
                result = func(*args, **kwargs)

                # Log success message if provided
                if log_msg:
                    logger.info(log_msg)
                
                # Handle different return types
                if isinstance(result, tuple):
                    return make_response(*result)
                return make_response(result, 200)

            except Exception as e:
                # check if exception is in custom exceptions
                # log as error and return 400 to handle in frontend
                if custom_exceptions:
                    for custom_exception in custom_exceptions:
                        if isinstance(e, custom_exception):
                            logger.error(e)
                            # Create a more structured error response
                            error_response = {
                                'error': e.message if hasattr(e, 'message') else str(e),
                                'type': e.__class__.__name__,
                                'route': route.value,
                                'context': {
                                    'file_type': current_file_type,
                                    'route': route.value,
                                }
                            }
                            return make_response(error_response, 400)
                        
                # log as critical and return 500 to handle in frontend
                logger.critical(
                    f"{error_msg}: {str(e)}", 
                    exc_info=True
                )
                
                # Create a more structured error response for critical errors
                error_response = {
                    'error': error_msg,
                    'details': str(e),
                    'type': e.__class__.__name__,
                    'route': route.value,
                    'context': {
                        'file_type': current_file_type,
                        'route': route.value,
                    }
                }
                return make_response(error_response, 500)
            
        return wrapper

    return decorator

# check string is numeric
def is_numeric(value):
    try:
        float(value)  # Try to convert to a number
        return True
    except (ValueError, TypeError):
        return False

def sanitize_array_for_json(arr: np.ndarray) -> List[List[float]]:
    """Convert numpy array to JSON-serializable format, replacing NaN with None.
    
    Parameters
    ----------
    arr : np.ndarray
        Input array that may contain NaN values
        
    Returns
    -------
    List[List[float]]
        JSON-serializable 2D list with NaN values replaced by None
    """
    # Replace NaN with None and convert to list
    return np.where(np.isnan(arr), None, arr).tolist()


# Function to convert strings (passed via fetch) to list of floats
def str_to_float_list(string):
    return list(map(float, string.split(',')))
